import os
import logging
import psutil
import numpy as np
import argparse
import traceback
import copy
import time
import glob
import sys
import os
import socket
import threading
from casatasks import casalog
from casatools import msmetadata
from datetime import datetime as dt
from multiprocessing import Process, Event
from meersolar.utils import *
from prefect import flow, task
from prefect.context import get_run_context
from prefect_dask.task_runners import DaskTaskRunner

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()

   
@task(name="making_dynamic_spectra",retries=2,retry_delay_seconds=10,log_prints=True)
def run_ds_jobs(
    msname,
    workdir,
    outdir,
    target_scans=[],
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Make dynamic spectra of the target scans

    Parameters
    ----------
    msname : str
        Name of the measurement set
    workdir : str
        Name of the work directory
    outdir : str
        Output directory
    target_scans : list, optional
        Target scans
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    ds_basename = "ds_targets"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{ds_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ##################
        print("###########################")
        print("Making dynamic spectra of target scans .....")
        print("###########################")
        ##########################
        # Making dynamic spectrum
        ##########################
        from meersolar.meerpipeline import meer_make_ds
        msg = meer_make_ds.main(
            msname,
            workdir,
            outdir,
            target_scans=target_scans,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            logfile=logfile,
            jobid=jobid,
            start_remote_log=remote_log,
            dask_addr=dask_addr,
        )
    finally:    
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Dynamic spectrum making is failed.")
    else:
        return msg


@task(name="attenuation_calibration",retries=2,retry_delay_seconds=10,log_prints=True)
def run_noise_diode_cal(
    msname,
    workdir,
    caldir,
    keep_backup=False,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Perform noise diode based flux calibration

    Parameters
    ----------
    msname: str
        Name of the measurement set
    workdir : str
        Working directory
    caldir : str
        Caltable directory
    keep_backup : bool, optional
        Keep backup
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message for noise diode based flux calibration
    """
    msname = msname.rstrip("/")
    noisecal_basename = "noise_cal"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{noisecal_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        #################
        print("###########################")
        print("Performing noise diode based flux calibration .....")
        print("###########################")
        #################
        # Attenuation calibration
        #################
        from meersolar.meerpipeline import do_fluxcal
        msg=do_fluxcal.main(
            msname,
            workdir,
            caldir,
            keep_backup=keep_backup,
            start_remote_log=remote_log,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            logfile=logfile,
            jobid=jobid,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Attenuation calibration is failed.")
    else:
        return msg
    

@task(name="partitioning_calibrator",retries=2,retry_delay_seconds=10,log_prints=True)
def run_partition(
    msname,
    workdir,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Perform basic calibration

    Parameters
    ----------
    msname: str
        Name of the measurement set
    workdir : str
        Working directory
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    msname = msname.rstrip("/")
    partition_basename = f"partition_cal"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{partition_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        msmd = msmetadata()
        msmd.open(msname)
        nchan = msmd.nchan(0)
        times = msmd.timesforfield(0)
        msmd.close()
        if len(times) == 1:
            timeres = msmd.exposuretime(scan)["value"]
        else:
            timeres = times[1] - times[0]
        if nchan > 1024:
            width = int(nchan / 1024)
            if width < 1:
                width = 1
        else:
            width = 1
        if timeres < 8:
            timebin = "8s"
        else:
            timebin = ""
        target_scans, cal_scans, f_scans, g_scans, p_scans = get_cal_target_scans(msname)
        cal_scans_copy = copy.deepcopy(cal_scans)
        for s in cal_scans:
            noise_cal_scan = determine_noise_diode_cal_scan(msname, s)
            if noise_cal_scan:
                print(f"Removing noise-diode scan: {s}")
                cal_scans_copy.remove(s)
        cal_scans = copy.deepcopy(cal_scans_copy)
        cal_scans = ",".join([str(s) for s in cal_scans])
        calibrator_ms = workdir + "/calibrator.ms"
        ##############
        print("###########################")
        print("Partitioning measurement set ...")
        print("###########################")
        ###################
        # Paritioning calibrator scans
        ###################
        from meersolar.meerpipeline import do_partition
        msg = do_partition.main(
            msname,
            outputms=calibrator_ms,
            workdir=workdir,
            scans=cal_scans,
            width=int(width),
            timebin=timebin,
            datacolumn="data",
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            logfile=logfile,
            jobid=jobid,
            start_remote_log=remote_log,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Partitioning calibrator scans is failed.")
    else:
        return msg


@task(name="spliting_target_scans",retries=2,retry_delay_seconds=10,log_prints=True)
def run_target_split_jobs(
    msname,
    workdir,
    datacolumn="data",
    spw="",
    timeres=-1,
    freqres=-1,
    target_freq_chunk=-1,
    n_spectral_chunk=-1,
    target_scans=[],
    prefix="targets",
    merge_spws=False,
    time_window=-1,
    time_interval=-1,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    max_cpu_frac=0.8,
    max_mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Split target scans

    Parameters
    ----------
    msname: str
        Name of the measurement set
    workdir : str
        Working directory
    datacolumn : str, optional
        Data column
    spw : str, optional
        Spectral window to split
    timeres : float, optional
        Time bin to average in seconds
    freqres : float, optional
        Frequency averaging in MHz
    target_freq_chunk : float, optional
        Target frequency chunk in MHz
    n_spectral_chunk : int, optional
        Number of spectral chunks to split
    target_scans : list, optional
        Target scans
    prefix : str, optional
        Prefix of splited targets
    merge_spws : bool, optional
        Merge spectral windows
    time_window : float, optional
        Time window in seconds
    time_interval : float, optional
        Time interval in seconds
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address
        
    Returns
    -------
    int
        Success message for spliting target scans
    """
    msname = msname.rstrip("/")
    split_basename = f"split_{prefix}"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{split_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ############
        print("###########################")
        print(f"Spliting {prefix} scans .....")
        print("###########################")
        ##################
        # Spliting target scans
        ##################
        scans=",".join([str(s) for s in target_scans])
        from meersolar.meerpipeline import do_target_split
        msg = do_target_split.main(
            msname,
            workdir=workdir,
            datacolumn=datacolumn,
            spw=spw,
            scans=scans,
            time_window=time_window,
            time_interval=time_interval,
            spectral_chunk=target_freq_chunk,
            n_spectral_chunk=n_spectral_chunk,
            freqres=freqres,
            timeres=timeres,
            prefix=prefix,
            merge_spws=merge_spws,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            max_cpu_frac=float(max_cpu_frac),
            max_mem_frac=float(max_mem_frac),
            logfile=logfile,
            jobid=jobid,
            start_remote_log=remote_log,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Spliting target scans is failed.")
    else:
        return msg

    
@task(name="flagging",retries=2,retry_delay_seconds=10,log_prints=True)
def run_flag(
    msname,
    workdir,
    flag_calibrators=True,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Run flagging jobs

    Parameters
    ----------
    msname: str
        Name of the measurement set
    workdir : str
        Working directory
    flag_calibrators : bool, optional
        Flag calibrator fields
    jobid : int, optional
        Job ID
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    msname = msname.rstrip("/")
    if flag_calibrators:
        flagdimension = "freqtime"
        flagfield_type = "cal"
    else:
        flagdimension = "freq"
        flagfield_type = "target"
    flag_basename = (
        f"flagging_{flagfield_type}_" + os.path.basename(msname).split(".ms")[0]
    )
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{flag_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ##############
        print("###########################")
        print("Flagging ....")
        print("###########################")
        ########################
        # Calibrator ms flagging
        ########################
        from meersolar.meerpipeline import flagging
        msg = flagging.main(
            msname,
            workdir=workdir,
            datacolumn="DATA",
            flag_bad_ants=True,
            flag_bad_spw=True,
            use_tfcrop=True,
            use_rflag=False,
            flag_autocorr=True,
            flagbackup=True,
            flagdimension=flagdimension,
            cpu_frac=float(cpu_frac),
            mem_frac=float(cpu_frac),
            logfile=logfile,
            jobid=jobid,
            start_remote_log=remote_log,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Calibrator flagging is failed.")
    else:
        return msg


@task(name="importing_model_visibilities",retries=2,retry_delay_seconds=10,log_prints=True)
def run_import_model(
    msname,
    workdir,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Importing calibrator models

    Parameters
    ----------
    msname: str
        Name of the measurement set
    workdir : str
        Working directory
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    msname = msname.rstrip("/")
    model_basename = "modeling_" + os.path.basename(msname).split(".ms")[0]
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{model_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ##############
        print("###########################")
        print("Importing model visibilities ....")
        print("###########################")
        ########################
        # Calibrator ms flagging
        ########################
        from meersolar.meerpipeline import import_model
        msg = import_model.main(
            msname,
            workdir=workdir,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            logfile=logfile,
            jobid=jobid,
            start_remote_log=remote_log,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Importing calibrator model is failed.")
    else:
        return msg
    
    
@task(name="basic_calibration",retries=2,retry_delay_seconds=10,log_prints=True)
def run_basic_cal_jobs(
    msname,
    workdir,
    caldir,
    perform_polcal=False,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    keep_backup=False,
    remote_log=False,
    dask_addr=None,
):
    """
    Perform basic calibration

    Parameters
    ----------
    msname: str
        Name of the measurement set
    workdir : str
        Working directory
    caldir : str
        Caltable directory
    perform_polcal : bool, optional
        Perform full polarization calibration
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    keep_backup : bool, optional
        Keep backups
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message for basic calibration
    """
    msname = msname.rstrip("/")
    cal_basename = "basic_cal"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{cal_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ##############
        print("###########################")
        print("Performing basic calibration .....")
        print("###########################")
        ########################
        # Basic calibration
        ########################
        from meersolar.meerpipeline import basic_cal
        msg = basic_cal.main(
            msname,
            workdir,
            caldir,
            perform_polcal=perform_polcal,
            keep_backup=keep_backup,
            start_remote_log=remote_log,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            logfile=logfile,
            jobid=jobid,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Basic calibration is failed.")
    else:
        return msg
    

@task(name="applying_basic_calibration",retries=2,retry_delay_seconds=10,log_prints=True)
def run_apply_basiccal_sol(
    target_mslist,
    workdir,
    caldir,
    use_only_bandpass=False,
    overwrite_datacolumn=True,
    applymode="calflag",
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Apply basic calibration solutions on splited target scans

    Parameters
    ----------
    target_mslist: list
        Target measurement set list
    workdir : str
        Working directory
    caldir : str
        Caltable directory
    use_only_bandpass : bool
        Use only bandpass solutions
    applymode : str, optional
        Applycal mode
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    overwrite_datacolumn : bool
        Overwrite data column or not
    remote_log: bool, optional
        Start remote logger
    dask_addr: str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message for applying calibration solutions and spliting target scans
    """
    mslist = ",".join(target_mslist)
    applycal_basename = "apply_basiccal"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{applycal_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ######################
        print("###########################")
        print("Applying basic calibration solutions on target scans .....")
        print("###########################")
        ######################
        # Applying basic calibration
        ######################
        from meersolar.meerpipeline import do_apply_basiccal
        msg = do_apply_basiccal.main(
            mslist,
            workdir,
            caldir,
            use_only_bandpass=use_only_bandpass,
            applymode=applymode,
            overwrite_datacolumn=overwrite_datacolumn,
            start_remote_log=remote_log,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            logfile=logfile,
            jobid=jobid,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Applying basic calibration solutions is failed.")
    else:
        return msg
  
        
@task(name="solar_sidereal_correction",retries=2,retry_delay_seconds=10,log_prints=True)
def run_solar_siderealcor_jobs(
    mslist,
    workdir,
    prefix="targets",
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    max_cpu_frac=0.8,
    max_mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Apply sidereal motion correction of the Sun

    Parameters
    ----------
    mslist: str
        List of the measurement sets
    workdir : str
        Work directory
    prefix : str, optional
        Measurement set prefix
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    max_cpu_frac : float, optional
        Maximum CPU fraction to use
    max_mem_frac : float, optional
        Maximum memory fraction to use
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    mslist = ",".join(mslist)
    sidereal_basename = f"cor_sidereal_{prefix}"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{sidereal_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        #######################
        print("###########################")
        print("Correcting sidereal motion .....")
        print("###########################")
        #######################
        # Sidereal motion correction
        #######################
        from meersolar.meerpipeline import do_sidereal_cor
        msg = do_sidereal_cor.main(
            mslist,
            workdir=workdir,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            max_cpu_frac=float(max_cpu_frac),
            max_mem_frac=float(max_mem_frac),
            logfile=logfile,
            jobid=jobid,
            start_remote_log=remote_log,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Solar sidereal motion correction is failed.")
    else:
        return msg
    
    
@task(name="selfcal",retries=2,retry_delay_seconds=10,log_prints=True)
def run_selfcal_jobs(
    mslist,
    workdir,
    caldir,
    start_thresh=5.0,
    stop_thresh=3.0,
    max_iter=100,
    max_DR=1000,
    min_iter=2,
    conv_frac=0.3,
    solint="60s",
    do_apcal=True,
    solar_selfcal=True,
    keep_backup=False,
    uvrange="",
    minuv=0,
    weight="briggs",
    robust=0.0,
    applymode="calonly",
    min_tol_factor=10.0,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Self-calibration on target scans

    Parameters
    ----------
    mslist: list
        Target measurement set list
    workdir : str
        Working directory
    caldir : str
        Caltable directory
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    start_threshold : int, optional
        Start CLEAN threhold
    end_threshold : int, optional
        End CLEAN threshold
    max_iter : int, optional
        Maximum numbers of selfcal iterations
    max_DR : float, optional
        Maximum dynamic range
    min_iter : int, optional
        Minimum numbers of seflcal iterations at different stages
    conv_frac : float, optional
        Dynamic range fractional change to consider as converged
    uvrange : str, optional
        UV-range for calibration
    minuv : float, optionial
        Minimum UV-lambda to use in imaging
    weight : str, optional
        Image weighitng scheme
    robust : float, optional
        Robustness parameter for briggs weighting
    solint : str, optional
        Solutions interval
    do_apcal : bool, optional
        Perform ap-selfcal or not
    min_tol_factor : float, optional
        Minimum tolerance in temporal variation in imaging
    applymode : str, optional
        Solution apply mode
    solar_selfcal : bool, optional
        Whether is is solar selfcal or not
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message for self-calibration
    """
    mslist = ",".join(mslist)
    selfcal_basename = "selfcal_targets"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{selfcal_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ########################
        print("###########################")
        print("Performing self-calibration of target scans .....")
        print("###########################")
        ########################
        # Selfcal jobs
        ########################
        from meersolar.meerpipeline import do_selfcal
        msg = do_selfcal.main(
            mslist,
            workdir,
            caldir,
            start_thresh=float(start_thresh),
            stop_thresh=float(stop_thresh),
            max_iter=float(max_iter),
            max_DR=float(max_DR),
            min_iter=float(min_iter),
            conv_frac=float(conv_frac),
            solint=solint,
            uvrange=uvrange,
            minuv=float(minuv),
            weight=weight,
            robust=float(robust),
            applymode=applymode,
            min_tol_factor=float(min_tol_factor),
            do_apcal=do_apcal,
            solar_selfcal=solar_selfcal,
            keep_backup=keep_backup,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            jobid=jobid,
            start_remote_log=remote_log,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Self-calibration is failed.")
    else:
        return msg
 

@task(name="applying_self-calibration",retries=2,retry_delay_seconds=10,log_prints=True)
def run_apply_selfcal_sol(
    target_mslist,
    workdir,
    caldir,
    overwrite_datacolumn=True,
    applymode="calflag",
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Apply self-calibration solutions on splited target scans

    Parameters
    ----------
    target_mslist: list
        Target measurement set list
    workdir : str
        Working directory
    caldir : str
        Caltable directory
    use_only_bandpass : bool
        Use only bandpass solutions
    applymode : str, optional
        Applycal mode
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    overwrite_datacolumn : bool
        Overwrite data column or not
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message for applying calibration solutions and spliting target scans
    """
    mslist = ",".join(target_mslist)
    applycal_basename = "apply_selfcal"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{applycal_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ##################
        print("###########################")
        print("Applying self-calibration solutions on target scans .....")
        print("###########################")
        ########################
        # Applying self-calibration
        ########################
        from meersolar.meerpipeline import do_apply_selfcal
        msg = do_apply_selfcal.main(
            mslist,
            workdir,
            caldir,
            applymode=applymode,
            overwrite_datacolumn=overwrite_datacolumn,
            start_remote_log=remote_log,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            logfile=logfile,
            jobid=jobid,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Applying self-calibration solutions is failed.")
    else:
        return msg
    
        
@task(name="imaging",retries=2,retry_delay_seconds=10,log_prints=True)
def run_imaging_jobs(
    mslist,
    workdir,
    outdir,
    freqrange="",
    timerange="",
    minuv=-1,
    weight="briggs",
    robust=0.0,
    pol="IQUV",
    freqres=-1,
    timeres=-1,
    band="",
    threshold=1.0,
    use_multiscale=True,
    use_solar_mask=True,
    cutout_rsun=2.5,
    make_overlay=True,
    savemodel=False,
    saveres=False,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Imaging on target scans

    Parameters
    ----------
    mslist: list
        Target measurement set list
    workdir : str
        Working directory
    outdir : str
        Output image directory
    freqrange : str, optional
        Frequency range to image in MHz
    timerange : str, optional
        Time range to image (YYYY/MM/DD/hh:mm:ss~YYYY/MM/DD/hh:mm:ss)
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    minuv : float, optionial
        Minimum UV-lambda to use in imaging
    weight : str, optional
        Imaging weighting
    robust : float, optional
        Briggs weighting robust parameter (-1 to 1)
    pol : str, optional
        Stokes parameters to image
    freqres : float, optional
        Frequency resolution of spectral chunk in MHz (default : -1, no spectral chunking)
    timeres : float, optional
        Time resolution of temporal chunks in MHz (default : -1, no temporal chunking)
    band : str, optional
        Band name
    threshold : float, optional
        CLEAN threshold
    use_multiscale : bool, optional
        Use multiscale or not
    use_solar_mask : bool, optional
        Use solar mask or not
    cutout_rsun : float, optional
        Cutout image size from center in solar radii (default : 2.5 solar radii)
    make_overlay : bool, optional
        Make SUVI MeerKAT overlay
    savemodel : bool, optional
        Save model images or not
    saveres : bool, optional
        Save residual images or not
    remote_log: bool, optional
        Start remote logger
    dask_addr : str, optional
        Dask scheduler address 

    Returns
    -------
    int
        Success message for imaging
    """
    mslist = ",".join(mslist)
    imaging_basename = "imaging_targets"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{imaging_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ######################
        print("###########################")
        print("Performing imaging of target scans .....")
        print("###########################")
        #######################
        # Performing imaging
        #######################
        from meersolar.meerpipeline import do_imaging
        msg = do_imaging.main(
            mslist,
            workdir,
            outdir,
            freqrange=freqrange,
            timerange=timerange,
            pol=pol,
            freqres=float(freqres),
            timeres=float(timeres),
            weight=weight,
            robust=float(robust),
            minuv=float(minuv),
            threshold=float(threshold),
            band=band,
            cutout_rsun=float(cutout_rsun),
            use_multiscale=use_multiscale,
            use_solar_mask=use_solar_mask,
            savemodel=savemodel,
            saveres=saveres,
            make_overlay=make_overlay,
            start_remote_log=remote_log,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            jobid=jobid,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Imaging is failed.")
    else:
        return msg
        
  
@task(name="applying_primary_beam",retries=2,retry_delay_seconds=10,log_prints=True)
def run_apply_pbcor(
    imagedir,
    workdir,
    apply_parang=True,
    jobid=0,
    cpu_frac=0.8,
    mem_frac=0.8,
    remote_log=False,
    dask_addr=None,
):
    """
    Apply primary beam corrections on all images

    Parameters
    ----------
    imagedir: str
        Image directory name
    workdir : str
        Work directory
    apply_parang : bool, optional
        Apply parallactic angle correction
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    remote_log: bool, optional
        Start remote logger
    dask_scheduler : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message for applying primary beam correction on all images
    """
    applypbcor_basename = "apply_pbcor"
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/{applypbcor_basename}.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    task_id=str(ctx.task_run.id)
    task_name = ctx.task_run.name
    stop_event = Event()
    log_thread = start_log_task_saver(task_id, task_name, logfile, poll_interval=3, stop_event=stop_event)
    try:
        ###################
        print("###########################")
        print("Applying primary beam corrections on all images .....")
        print("###########################")
        #####################
        # Applying primary beam correction
        #####################
        from meersolar.meerpipeline import meer_pbcor
        msg = meer_pbcor.main(
            imagedir,
            workdir=workdir,
            apply_parang=apply_parang,
            cpu_frac=float(cpu_frac),
            mem_frac=float(mem_frac),
            logfile=logfile,
            jobid=jobid,
            start_remote_log=remote_log,
            dask_addr=dask_addr,
        )
    finally:
        stop_event.set()
        log_thread.join(timeout=5)
    if msg!=0:
        raise RuntimeError("Primary beam correction is failed.")
    else:
        return msg
   

@flow(name="MeerSOLAR Master control",version="3.0",description="Calibration and Imaging Pipeline for MeerKAT Solar Observation",log_prints=True)
def master_control(
    msname,
    workdir,
    outdir,
    solar_data=True,
    # Pre-calibration
    do_forcereset_weightflag=False,
    do_cal_partition=True,
    do_cal_flag=True,
    do_import_model=True,
    # Basic calibration
    do_basic_cal=True,
    do_noise_cal=True,
    do_applycal=True,
    # Target data preparation
    do_target_split=True,
    target_scans=[],
    freqrange="",
    timerange="",
    uvrange="",
    # Polarization calibration
    do_polcal=False,
    # Self-calibration
    do_selfcal=True,
    do_selfcal_split=True,
    do_apply_selfcal=True,
    do_ap_selfcal=True,
    solar_selfcal=True,
    solint="5min",
    # Sidereal correction
    do_sidereal_cor=False,
    # Dynamic spectra
    make_ds=True,
    # Imaging
    do_imaging=True,
    do_pbcor=True,
    weight="briggs",
    robust=0.0,
    minuv=0,
    image_freqres=-1,
    image_timeres=-1,
    pol="IQUV",
    apply_parang=True,
    clean_threshold=1.0,
    use_multiscale=True,
    use_solar_mask=True,
    cutout_rsun=2.5,
    make_overlay=True,
    # Resource settings
    cpu_frac=0.8,
    mem_frac=0.8,
    n_nodes=0,
    keep_backup=False,
    # Remote logging
    remote_logger=False,
    dask_addr=None,
    jobid=None,
):
    """
    Master controller of the entire pipeline

    Parameters
    ----------
    msname : str
        Measurement set name
    workdir : str
        Work directory path
    outdir : str
        Output directory
    solar_data : bool, optional
        Whether it is solar data or not

    do_forcereset_weightflag : bool, optional
        Reset weights and flags of the input ms
    do_cal_partition : bool, optional
        Make calibrator multi-MS
    do_cal_flag : bool, optional
        Perform flagging on calibrator
    do_import_model : bool, optional
        Import model visibilities of flux and polarization calibrators

    do_basic_cal : bool, optional
        Perform basic calibration
    do_noise_cal : bool, optional
        Peform calibration of solar attenuators using noise diode (only used if solar_data=True)
    do_applycal : bool, optional
        Apply basic calibration on target scans

    do_target_split : bool, optional
        Split target scans into chunks
    target_scans : list, optional
        Target scans to self-cal and image
    freqrange : str, optional
        Frequency range to image in MHz (xx1~xx2,xx3~xx4,)
    timerange : str, optional
        Time range to image in YYYY/MM/DD/hh:mm:ss format (tt1~tt2,tt3~tt4,...)
    uvrange : str, optional
        UV-range for calibration

    do_polcal : bool, optional
        Perform full-polarization calibration and imaging

    do_selfcal : bool, optional
        Perform self-calibration
    do_selfcal_split : bool, optional
        Split data after each round of self-calibration
    do_apply_selfcal : bool, optional
        Apply self-calibration solutions
    do_ap_selfcal : bool, optional
        Perform amplitude-phase self-cal or not
    solar_selfcal : bool, optional
        Whether self-calibration is performing on solar observation or not
    solint : str, optional
        Solution intervals in self-cal

    do_sidereal_cor : bool, optional
        Perform solar sidereal motion correction or not

    make_ds : bool, optional
        Make dynamic spectra

    do_imaging : bool, optional
        Perform final imaging
    do_pbcor : bool, optional
        Perform primary beam correction
    weight : str, optional
        Image weighting
    robust : float, optional
        Robust parameter for briggs weighting (-1 to 1)
    minuv : float, optional
        Minimum UV-lambda for final imaging
    image_freqres : float, optional
        Image frequency resolution in MHz (-1 means full bandwidth)
    image_timeres : float, optional
        Image temporal resolution in seconds (-1 means full scan duration)
    pol : str, optional
        Stokes parameters of final imaging
    apply_parang : bool, optional
        Apply parallactic angle correction
    clean_threshold : float, optional
        CLEAN threshold of final imaging
    use_multiscale : bool, optional
        Use multiscale scales or not
    use_solar_mask : bool, optional
        Use solar mask
    cutout_rsun : float, optional
        Cutout image size from center in solar radii (default : 2.5 solar radii)
    make_overlay : bool, optional
        Make SUVI MeerKAT overlay

    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    n_nodes: int, optional
        Number of nodes to use (Only for cluster architecture)
    keep_backup : bool, optional
        Keep backup of self-cal rounds and final models and residual images

    remote_logger : bool, optional
        Enable remote logging of the pipeline status
    dask_addr : str, optional
        Dask scheduler address
    jobid : str, optional
        Job ID

    Returns
    -------
    int
        Success message
    """
    ###################################
    # Preparing working directories
    ###################################
    print("Preparing working directories....")
    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    workdir = workdir.rstrip("/")
    
    #################################
    # Setup logger
    #################################
    logdir = f"{workdir}/logs"
    os.makedirs(logdir, exist_ok=True)
    logfile = f"{logdir}/main.log"
    if os.path.exists(logfile):
        os.remove(logfile)
    ctx = get_run_context()
    flow_id=str(ctx.flow_run.id)
    flow_name = ctx.flow_run.name
    stop_event = Event()
    log_thread = start_flow_log_saver(flow_id, flow_name, logfile, poll_interval=3, stop_event=stop_event)
    
    #####################################
    # Initiating meersolar data
    #####################################
    from meersolar.meerpipeline.init_data import init_meersolar_data
    init_meersolar_data()
    
    ###################################################
    # Measurement set check and other working directory
    ###################################################
    msname = os.path.abspath(msname.rstrip("/"))
    if os.path.exists(msname) == False:
        print("Please provide a valid measurement set location.")
        return 1
    valid_ms=check_datacolumn_valid(msname)
    if valid_ms is not True:
        print (f"Measurement set : {msname} is corrupted.")
        return 1
    mspath = os.path.dirname(msname)
    if outdir == "":
        outdir = workdir
    outdir = outdir.rstrip("/")
    caldir = outdir + "/caltables"
    caldir = caldir.rstrip("/")
    os.makedirs(workdir, exist_ok=True)
    os.makedirs(outdir, exist_ok=True)
    os.makedirs(caldir, exist_ok=True)
    
    try:
        ####################################
        # Job and process IDs
        ####################################
        pid = os.getpid()
        if jobid is None:
            jobid = get_jobid()
        main_job_file = save_main_process_info(
            pid,
            jobid,
            os.path.abspath(msname),
            os.path.abspath(workdir),
            os.path.abspath(outdir),
            cpu_frac,
            mem_frac,
        )
        print("###########################")
        print(f"MeerSOLAR Job ID: {jobid}")
        print(f"Work directory: {workdir}")
        print(f"Final product directory: {outdir}")
        print("###########################")
        #####################################
        # Moving into work directory
        #####################################
        os.chdir(workdir)
        if remote_logger:
            trial=0
            while trial<=5:
                remote_link = get_remote_logger_link()
                if remote_link != "":
                    break
                else:
                    time.sleep(5)
                    trial+=1
            if remote_link=="":
                print("Please provide a valid remote link.")
                remote_logger = False

        if not remote_logger:
            emails = get_emails()
            timestamp = dt.utcnow().strftime("%Y-%m-%dT%H:%M:%S")
            if emails != "":
                email_subject = f"MeerSOLAR Logger Details: {timestamp}"

                email_msg = (
                    f"MeerSOLAR user,\n\n"
                    f"MeerSOLAR Job ID: {jobid}\n\n"
                    f"Best,\n"
                    f"MeerSOLAR"
                )
                from meersolar.data.sendmail import (
                    send_paircars_notification as send_notification,
                )

                success_msg, error_msg = send_notification(
                    emails, email_subject, email_msg
                )
        else:
            ####################################
            # Job name and logging password
            ####################################
            hostname = socket.gethostname()
            timestamp = dt.utcnow().strftime("%Y-%m-%dT%H:%M:%S")
            job_name = f"{hostname} :: {timestamp} :: {os.path.basename(msname).split('.ms')[0]}"
            timestamp1 = dt.utcnow().strftime("%Y%m%dT%H%M%S")
            remote_job_id = (
                f"{hostname}_{timestamp1}_{os.path.basename(msname).split('.ms')[0]}"
            )
            password = generate_password()
            np.save(
                f"{workdir}/jobname_password.npy",
                np.array([job_name, password], dtype="object"),
            )
            print(
                "############################################################################"
            )
            print(remote_link)
            print(f"Job ID: {job_name}")
            print(f"Remote access password: {password}")
            print(
                "#############################################################################"
            )
            emails = get_emails()
            if emails != "":
                email_subject = f"MeerSOLAR Logger Details: {timestamp}"

                email_msg = (
                    f"MeerSOLAR user,\n\n"
                    f"MeerSOLAR Job ID: {jobid}\n\n"
                    f"Remote logger Job ID: {job_name}\n"
                    f"Remote access password: {password}\n\n"
                    f"Best,\n"
                    f"MeerSOLAR"
                )
                from meersolar.data.sendmail import (
                    send_paircars_notification as send_notification,
                )

                success_msg, error_msg = send_notification(
                    emails, email_subject, email_msg
                )

        #####################################
        # Settings for solar data
        #####################################
        if solar_data:
            if not use_solar_mask:
                print("Use solar mask during CLEANing.")
                use_solar_mask = True
            if not solar_selfcal:
                solar_selfcal = True
            full_FoV = False
        else:
            if do_noise_cal:
                print(
                    "Turning off noise diode based calibration for non-solar observation."
                )
                do_noise_cal = False
            if use_solar_mask:
                print("Stop using solar mask during CLEANing.")
                use_solar_mask = False
            if solar_selfcal:
                solar_selfcal = False
            full_FoV = True

        ##################################################
        # Target spliting spectral and temporal chunks
        ##################################################
        if image_timeres > (2 * 3660):  # If more than 2 hours
            print(
                f"Image time integration is more than 2 hours, which may cause smearing due to solar differential rotation."
            )

        #####################################################################
        # Checking if ms is full pol for polarization calibration and imaging
        #####################################################################
        if do_polcal:
            print(
                "Checking measurement set suitability for polarization calibration...."
            )
            msmd = msmetadata()
            msmd.open(msname)
            npol = msmd.ncorrforpol()[0]
            msmd.close()
            if npol < 4:
                print(
                    "Measurement set is not full-polar. Do not performing polarization analysis."
                )
                do_polcal = False

        #################################################
        # Determining maximum allowed frequency averaging
        #################################################
        print("Estimating optimal frequency averaging....")
        max_freqres = calc_bw_smearing_freqwidth(msname, full_FoV=full_FoV)
        if image_freqres > 0:
            freqavg = round(min(image_freqres, max_freqres), 1)
        else:
            freqavg = round(max_freqres, 1)

        ################################################
        # Determining maximum allowed temporal averaging
        ################################################
        print("Estimating optimal temporal averaging....")
        if solar_data:  # For solar data, it is assumed Sun is tracked.
            max_timeres = calc_time_smearing_timewidth(msname)
        else:
            max_timeres = min(
                calc_time_smearing_timewidth(msname), max_time_solar_smearing(msname)
            )
        if image_timeres > 0:
            timeavg = round(min(image_timeres, max_timeres), 1)
        else:
            timeavg = round(max_timeres, 1)

        #########################################
        # Target ms frequency chunk based on band
        #########################################
        print("Determing bad spectral channels....")
        bad_spws = get_bad_chans(msname)
        if bad_spws != "":
            bad_spws = bad_spws.split("0:")[-1].split(";")
            good_start = []
            good_end = []
            for i in range(len(bad_spws) - 1):
                start_chan = int(bad_spws[i].split("~")[-1]) + 1
                end_chan = int(bad_spws[i + 1].split("~")[0]) - 1
                good_start.append(start_chan)
                good_end.append(end_chan)
            start_chan = min(good_start)
            end_chan = max(good_end)
        else:
            msmd = msmetadata()
            msmd.open(msname)
            nchan = msmd.nchan(0)
            msmd.close()
            start_chan = 0
            end_chan = nchan
        spw = f"0:{start_chan}~{end_chan}"

        #############################################################
        # Determining numbers of spectral chunks for parallel imaging
        #############################################################
        print("Determining spectral chunks for parallel imaging....")
        if image_freqres < 0:
            target_freq_chunk = -1
            nchunk = 1
        else:
            msmd = msmetadata()
            msmd.open(msname)
            chanres = msmd.chanres(0, unit="MHz")[0]
            msmd.close()
            total_bw = chanres * (end_chan - start_chan)
            nchunk = int(total_bw / image_freqres)
            if nchunk > max(
                4, n_nodes
            ):  # Maximum 4 chunking or number of compute nodes
                nchunk = max(4, n_nodes)
                target_freq_chunk = total_bw / nchunk
            else:
                nchnuk = 1
                target_freq_chunk = image_freqres

        #############################
        # Reset any previous weights
        ############################
        print("Resetting previous flags and weights....")
        cpu_usage = psutil.cpu_percent(interval=1)  # Average over 1 second
        total_cpus = psutil.cpu_count(logical=True)
        available_cpus = int(total_cpus * (1 - cpu_usage / 100.0))
        available_cpus = max(1, available_cpus)  # Avoid zero workers
        reset_weights_and_flags(
            msname, n_threads=available_cpus, force_reset=do_forcereset_weightflag
        )
        
        #######################################
        # Run dynamic spectra making
        #######################################
        future_maskms=None
        if make_ds:
            future_maskms = run_ds_jobs.with_options(task_run_name=f"making_dynamic_spectra_{jobid}").submit(
                msname,
                workdir,
                outdir,
                jobid=jobid,
                target_scans=target_scans,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            msg=future_maskms.result()
            try:
                msg = future_maskms.result()
            except Exception as e:
                print("!!! WARNING : Error in making dynamic spectra. !!!")
                traceback.print_exc()
                msg = None
               
        ########################################
        # Run noise-diode based flux calibration
        ########################################
        future_noisecal=None
        if do_noise_cal:
            future_noisecal = run_noise_diode_cal.with_options(task_run_name=f"attenuation_calibration_{jobid}").submit(
                msname,
                workdir,
                caldir,
                jobid=jobid,
                keep_backup=keep_backup,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )  
            try:
                msg = future_noisecal.result()
                if os.path.exists(f"{workdir}/.noattcal"):
                    os.system(f"rm -rf {workdir}/.noattcal")
                os.system(f"touch {workdir}/.attcal")
            except Exception as e:
                print(
                    "!!!! WARNING: Error in running noise-diode based flux calibration. Flux density calibration may not be correct. !!!!"
                )
                traceback.print_exc()
                msg = None
                if os.path.exists(f"{workdir}/.attcal"):
                    os.system(f"rm -rf {workdir}/.attcal")
                os.system(f"touch {workdir}/.noattcal")     
        else:
            if os.path.exists(f"{workdir}/.attcal") == False:
                os.system(f"touch {workdir}/.noattcal")
            
        ##############################
        # Run partitioning jobs
        ##############################
        # If do partition or calibrator ms is not present in case of basic
        # calibration is requested
        calibrator_msname = workdir + "/calibrator.ms"
        future_partition = None
        if do_basic_cal and (
            do_cal_partition or os.path.exists(calibrator_msname) == False
        ):
            future_partition = run_partition.with_options(task_run_name=f"partitioning_calibrator_{jobid}").submit(
                msname,
                workdir,
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            try:
                msg = future_partition.result()
            except Exception as e:
                print("!!!! WARNING: Error in partitioning calibrator fields. !!!!")
                traceback.print_exc()
                return 1
                 
        ############################################
        # Spliting for self-cals
        ############################################
        # Spliting only if self-cal is requested
        if not do_selfcal_split and do_selfcal:
            selfcal_target_mslist = glob.glob(workdir + "/selfcals_scan*.ms")
            if len(selfcal_target_mslist) == 0:
                print(
                    "No measurement set is present for self-calibration. Spliting them.."
                )
                do_selfcal_split = True
        future_selfcal_split=None
        if do_selfcal and do_selfcal_split:
            prefix = "selfcals"
            try:
                time_interval = float(solint)
            except BaseException:
                if "s" in solint:
                    time_interval = float(solint.split("s")[0])
                elif "min" in solint:
                    time_interval = float(solint.split("min")[0]) * 60
                elif solint == "int":
                    time_interval = image_timeres
                else:
                    time_interval = -1
            future_selfcal_split = run_target_split_jobs.with_options(task_run_name=f"spliting_{prefix}_scans_{jobid}").submit(
                msname,
                workdir,
                datacolumn="data",
                freqres=freqavg,
                timeres=timeavg,
                target_freq_chunk=25,
                n_spectral_chunk=nchunk,  # Number of target spectral chunk
                target_scans=target_scans,
                prefix=prefix,
                merge_spws=True,
                time_window=min(60, time_interval),
                time_interval=time_interval,
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                max_cpu_frac=round(cpu_frac, 2),
                max_mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            ) 
            
        ##################################
        # Run flagging jobs on calibrators
        ##################################
        # Only if basic calibration is requested
        future_flag=None
        if do_cal_flag and do_basic_cal:
            if os.path.exists(calibrator_msname) == False:
                print(f"Calibrator ms: {calibrator_ms} is not present.")
                return 1
            future_flag = run_flag.with_options(task_run_name=f"flagging_{jobid}").submit(
                calibrator_msname,
                workdir,
                flag_calibrators=True,
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            try:
                msg = future_flag.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Flagging error. Examine calibration solutions with caution. !!!!"
                )
                traceback.print_exc()
                msg=1
          
        #################################
        # Import model
        #################################
        # Only if basic calibration is requested
        future_import_model=None
        if do_import_model and do_basic_cal:
            if os.path.exists(calibrator_msname) == False:
                print(f"Calibrator ms: {calibrator_ms} is not present.")
                return 1
            fluxcal_fields, fluxcal_scans = get_fluxcals(calibrator_msname)
            phasecal_fields, phasecal_scans, phasecal_fluxes = get_phasecals(
                calibrator_msname
            )
            calibrator_field = fluxcal_fields + phasecal_fields
            future_import_model = run_import_model.with_options(task_run_name=f"importing_model_visibilities_{jobid}").submit(
                calibrator_msname,
                workdir,
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            ) 
            try:
                msg = future_import_model.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Error in importing calibrator models. Not continuing calibration. !!!!"
                )
                traceback.print_exc()
                return 1

        ###############################
        # Run basic calibration
        ###############################
        use_only_bandpass = False
        future_basical=None
        if do_basic_cal:
            if os.path.exists(calibrator_msname) == False:
                print(f"Calibrator ms: {calibrator_ms} is not present.")
                return 1
            future_basical = run_basic_cal_jobs.with_options(task_run_name=f"basic_calibration_{jobid}").submit(
                calibrator_msname,
                workdir,
                caldir,
                perform_polcal=do_polcal,
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                keep_backup=keep_backup,
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )  
            try:
                msg = future_basical.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Error in basic calibration. Not continuing further. !!!!"
                )
                traceback.print_exc()
                return 1
           
        ##########################################
        # Checking presence of necessary caltables
        ##########################################
        if len(glob.glob(caldir + "/*.bcal")) == 0:
            print(f"No bandpass table is present in calibration directory : {caldir}.")
            return 1
        if len(glob.glob(caldir + "/*.gcal")) == 0:
            print(
                f"No time-dependent gaintable is present in calibration directory : {caldir}. Applying only bandpass solutions."
            )
            use_only_bandpass = True
            
        ############################################
        # Check if spliting for selfcal is fininshed
        ############################################
        if future_selfcal_split is not None:
            print("Checking spliting of target scans for selfcal status...")
            try:
                msg = future_selfcal_split.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Error in running spliting target scans for selfcal. !!!!"
                )
                do_selfcal = False
                traceback.print_exc()
                msg=1
                
        #############################################
        # Spliting target scans 
        #############################################
        # If corrected data is requested or imaging is requested
        prefix = "targets"
        future_split=None
        if do_target_split and (do_applycal or do_imaging):
            future_split = run_target_split_jobs.with_options(task_run_name=f"spliting_{prefix}_scans_{jobid}").submit(
                msname,
                workdir,
                datacolumn="data",
                spw=spw,
                target_freq_chunk=target_freq_chunk,
                freqres=freqavg,
                timeres=timeavg,
                n_spectral_chunk=-1,
                target_scans=target_scans,
                prefix=prefix,
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                max_cpu_frac=round(cpu_frac, 2),
                max_mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            
        ####################################
        # Filtering any corrupted ms
        #####################################
        selfcal_target_mslist = glob.glob(workdir + "/selfcals_scan*.ms")
        if (selfcal_target_mslist)==0:
            print(
                "!!!! WARNING: Error in running spliting target scans for selfcal. !!!!"
            )
            do_selfcal=False 
        if do_selfcal:
            print("Checking measurement sets before spawning self-calibrations....")
            filtered_mslist = []  # Filtering in case any ms is corrupted
            for ms in selfcal_target_mslist:
                checkcol = check_datacolumn_valid(ms)
                if checkcol:
                    filtered_mslist.append(ms)
                else:
                    print(f"Issue in : {ms}")
                    os.system(f"rm -rf {ms}")
            selfcal_mslist = filtered_mslist
            if len(selfcal_mslist) == 0:
                print(
                    "No splited target scan ms are available in work directory for selfcal. Not continuing further for selfcal."
                )
                do_selfcal = False
            print(f"Selfcal mslist : {[os.path.basename(i) for i in selfcal_mslist]}")
                        
        #########################################################
        # Applying solutions on target scans for self-calibration
        #########################################################
        future_apply_basical_selfcal=None
        if do_selfcal:  
            future_apply_basical_selfcal = run_apply_basiccal_sol.with_options(task_run_name=f"applying_basiccal_selfcal_{jobid}").submit(
                selfcal_mslist,
                workdir,
                caldir,
                use_only_bandpass=use_only_bandpass,
                overwrite_datacolumn=False,
                applymode="calflag",
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            try:
                msg = future_apply_basical_selfcal.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Error in applying basic calibration solutions on target scans. Not continuing further for selfcal.!!!!"
                )
                do_selfcal = False
                traceback.print_exc()
                msg=1
          
        ########################################
        # Performing self-calibration
        ########################################
        future_selfcal=None
        if do_selfcal:
            os.system(
                "rm -rf " + workdir + "/*selfcal " + workdir + "/caltables/*selfcal*"
            )
            future_sidereal_cor_selfcal=None
            if do_sidereal_cor:
                future_sidereal_cor_selfcal = run_solar_siderealcor_jobs.with_options(task_run_name=f"solar_sidereal_correction_{jobid}").submit(
                    selfcal_mslist,
                    workdir,
                    prefix="selfcals",
                    jobid=jobid,
                    cpu_frac=round(cpu_frac, 2),
                    mem_frac=round(mem_frac, 2),
                    max_cpu_frac=round(cpu_frac, 2),
                    max_mem_frac=round(mem_frac, 2),
                    remote_log=remote_logger,
                    dask_addr=dask_addr,
                )
                msg=future_sidereal_cor_selfcal.result()
                if msg != 0:
                    print("Sidereal correction is not successful.")
            future_selfcal = run_selfcal_jobs.with_options(task_run_name=f"selfcal_{jobid}").submit(
                selfcal_mslist,
                workdir,
                caldir,
                solint=solint,
                do_apcal=do_ap_selfcal,
                solar_selfcal=solar_selfcal,
                keep_backup=keep_backup,
                uvrange=uvrange,
                weight="briggs",
                robust=0.0,
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            msg=future_selfcal.result()
            try:
                msg = future_selfcal.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Error in self-calibration on target scans. Not applying self-calibration. !!!!"
                )
                do_apply_selfcal = False
                traceback.print_exc()
                msg=1
            
        ########################################
        # Checking self-cal caltables
        ########################################
        if do_imaging or do_apply_selfcal:
            selfcal_tables = glob.glob(caldir + "/*selfcal*")
            if len(selfcal_tables) == 0:
                print(
                    "Self-calibration is not performed and no self-calibration caltable is available."
                )
                do_apply_selfcal = False

        #############################################
        # Checking finishing of spliting target scans
        #############################################
        if future_split is not None:
            print("Checking spliting of target scans status...")
            try:
                msg = future_split.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Error in spliting target scans. !!!!"
                )
                traceback.print_exc()
                return 1
     
        if do_imaging or do_applycal or do_apply_selfcal:
            target_mslist = glob.glob(workdir + "/targets_scan*.ms")
            if len(target_mslist)==0:
                print(
                        "!!!! WARNING: No target scans are present. !!!!"
                    )    
                return 1

            ####################################
            # Filtering any corrupted ms
            #####################################
            print(
                "Checking final valid measurement sets before applying solutions and spawning imaging...."
            )
            filtered_mslist = []  # Filtering in case any ms is corrupted
            for ms in target_mslist:
                checkcol = check_datacolumn_valid(ms)
                if checkcol:
                    filtered_mslist.append(ms)
                else:
                    print(f"Issue in : {ms}")
                    os.system("rm -rf {ms}")
            target_mslist = filtered_mslist
            if len(target_mslist) == 0:
                print("No filtered target scan ms are available in work directory.")
                return 1

            if do_applycal or do_imaging:
                print(
                    f"Target scan mslist : {[os.path.basename(i) for i in target_mslist]}"
                )

        #########################################################
        # Applying basic solutions on target scans
        #########################################################
        future_apply_basical=None
        if do_applycal:
            future_apply_basical = run_apply_basiccal_sol.with_options(task_run_name=f"applying_basiccal_target_{jobid}").submit(
                target_mslist,
                workdir,
                caldir,
                use_only_bandpass=use_only_bandpass,
                overwrite_datacolumn=True,
                applymode="calflag",
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            try:
                msg = future_apply_basical.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Error in applying basic calibration solutions on target scans. Not continuing further.!!!!"
                )
                traceback.print_exc()
                return 1
            future_sidereal_cor=None
            if do_sidereal_cor:
                future_sidereal_cor = run_solar_siderealcor_jobs.with_options(task_run_name=f"solar_sidereal_correction_{jobid}").submit(
                    target_mslist,
                    workdir,
                    prefix="targets",
                    jobid=jobid,
                    cpu_frac=round(cpu_frac, 2),
                    mem_frac=round(mem_frac, 2),
                    max_cpu_frac=round(cpu_frac, 2),
                    max_mem_frac=round(mem_frac, 2),
                    remote_log=remote_logger,
                    dask_addr=dask_addr,
                )
                try:
                    msg = future_sidereal_cor.result()
                except Exception as e:
                    print(
                        "!!!! WARNING: Error in applying sidereal correction.!!!!"
                    )
                    traceback.print_exc()
                    msg=1

        ########################################
        # Apply self-calibration
        ########################################
        future_apply_selfcal=None
        if do_apply_selfcal:
            target_mslist = sorted(target_mslist)
            future_apply_selfcal = run_apply_selfcal_sol.with_options(task_run_name=f"applying_selfcal_{jobid}").submit(
                target_mslist,
                workdir,
                caldir,
                overwrite_datacolumn=False,
                applymode="calonly",
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            try:
                msg = future_apply_selfcal.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Error in applying self-calibration solutions on target scans. !!!!"
                )
                traceback.print_exc()
                msg=1
            
        #####################################
        # Imaging
        ######################################
        future_imaging=None
        if do_imaging:
            if (
                do_polcal == False
            ):  # Only if do_polcal is False, overwrite to make only Stokes I
                pol = "I"
            band = get_band_name(target_mslist[0])
            future_imaging = run_imaging_jobs.with_options(task_run_name=f"imaging_{jobid}").submit(
                target_mslist,
                workdir,
                outdir,
                freqrange=freqrange,
                timerange=timerange,
                minuv=minuv,
                weight=weight,
                robust=float(robust),
                pol=pol,
                band=band,
                freqres=image_freqres,
                timeres=image_timeres,
                threshold=float(clean_threshold),
                use_multiscale=use_multiscale,
                use_solar_mask=use_solar_mask,
                cutout_rsun=cutout_rsun,
                make_overlay=make_overlay,
                savemodel=keep_backup,
                saveres=keep_backup,
                jobid=jobid,
                cpu_frac=round(cpu_frac, 2),
                mem_frac=round(mem_frac, 2),
                remote_log=remote_logger,
                dask_addr=dask_addr,
            )
            try:
                msg = future_imaging.result()
            except Exception as e:
                print(
                    "!!!! WARNING: Final imaging on all measurement sets is not successful. Check the image directory. !!!!"
                )
                traceback.print_exc()
                return 1

        ###########################
        # Primary beam correction
        ###########################
        future_pbcor=None
        if do_pbcor:
            if weight == "briggs":
                weight_str = f"{weight}_{robust}"
            else:
                weight_str = weight
            if image_freqres == -1 and image_timeres == -1:
                imagedir = outdir + f"/imagedir_f_all_t_all_w_{weight_str}"
            elif image_freqres != -1 and image_timeres == -1:
                imagedir = (
                    outdir
                    + f"/imagedir_f_{round(float(image_freqres),1)}_t_all_w_{weight_str}"
                )
            elif image_freqres == -1 and image_timeres != -1:
                imagedir = (
                    outdir
                    + f"/imagedir_f_all_t_{round(float(image_timeres),1)}_w_{weight_str}"
                )
            else:
                imagedir = (
                    outdir
                    + f"/imagedir_f_{round(float(image_freqres),1)}_t_{round(float(image_timeres),1)}_w_{weight_str}"
                )
            imagedir = imagedir + "/images"
            images = glob.glob(imagedir + "/*.fits")
            if len(images) == 0:
                print(f"No image is present in image directory: {imagedir}")
            else:
                future_pbcor = run_apply_pbcor.with_options(task_run_name=f"applying_primary_beam_{jobid}").submit(
                    imagedir,
                    workdir,
                    apply_parang=apply_parang,
                    jobid=jobid,
                    cpu_frac=round(cpu_frac, 2),
                    mem_frac=round(mem_frac, 2),
                    remote_log=remote_logger,
                    dask_addr=dask_addr,
                )
                msg=future_pbcor.result()
                try:
                    msg = future_pbcor.result()
                except Exception as e:
                    print(
                        "!!!! WARNING: Primary beam corrections of the final images are not successful. !!!!"
                    )
                    traceback.print_exc()
                    return 1
                print(f"Final image directory: {os.path.dirname(imagedir)}")
                
        ###########################################
        # Successful exit
        ###########################################
        print(
            f"Calibration and imaging pipeline is successfully run on measurement set : {msname}"
        )
        return 0
    except Exception as e:
        traceback.print_exc()
        return 1
    finally:
        drop_cache(msname)
        drop_cache(workdir)
        drop_cache(outdir)
        stop_event.set()
        log_thread.join(timeout=5)

def cli():
    parser = argparse.ArgumentParser(
        description="Run MeerSOLAR for calibration and imaging of solar observations.",
        formatter_class=SmartDefaultsHelpFormatter,
    )
    # === Essential parameters ===
    essential = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    essential.add_argument("msname", type=str, help="Measurement set name")
    essential.add_argument(
        "--workdir",
        type=str,
        dest="workdir",
        required=True,
        help="Working directory",
    )
    essential.add_argument(
        "--outdir",
        type=str,
        dest="outdir",
        required=True,
        help="Output products directory",
    )

    # === Advanced calibration parameters ===
    advanced_cal = parser.add_argument_group(
        "###################\nAdvanced calibration parameters\n###################"
    )
    advanced_cal.add_argument(
        "--solint",
        type=str,
        default="5min",
        help="Solution interval for calibration (e.g. 'int', '10s', '5min', 'inf')",
    )
    advanced_cal.add_argument(
        "--cal_uvrange",
        type=str,
        default="",
        help="UV range to filter data for calibration (e.g. '>100klambda', '100~10000lambda')",
    )
    advanced_cal.add_argument(
        "--no_polcal",
        action="store_false",
        dest="do_polcal",
        help="Disable polarization calibration",
    )

    # === Advanced imaging and calibration parameters ===
    advanced_image = parser.add_argument_group(
        "###################\nAdvanced imaging parameters\n###################"
    )
    advanced_image.add_argument(
        "--target_scans",
        nargs="*",
        type=str,
        default=[],
        help="List of target scans to process (space-separated, e.g. 3 5 7)",
    )
    advanced_image.add_argument(
        "--freqrange",
        type=str,
        default="",
        help="Frequency range in MHz to select during imaging (comma-seperate, e.g. '100~110,130~140')",
    )
    advanced_image.add_argument(
        "--timerange",
        type=str,
        default="",
        help="Time range to select during imaging (comma-seperated, e.g. '2014/09/06/09:30:00~2014/09/06/09:45:00,2014/09/06/10:30:00~2014/09/06/10:45:00')",
    )
    advanced_image.add_argument(
        "--image_freqres",
        type=int,
        default=-1,
        help="Output image frequency resolution in MHz (-1 = full)",
    )
    advanced_image.add_argument(
        "--image_timeres",
        type=int,
        default=-1,
        help="Output image time resolution in seconds (-1 = full)",
    )
    advanced_image.add_argument(
        "--pol",
        type=str,
        default="IQUV",
        help="Stokes parameter(s) to image (e.g. 'I', 'XX', 'RR', 'IQUV')",
    )
    advanced_image.add_argument(
        "--minuv",
        type=float,
        default=0,
        help="Minimum baseline length (in wavelengths) to include in imaging",
    )
    advanced_image.add_argument(
        "--weight",
        type=str,
        default="briggs",
        help="Imaging weighting scheme (e.g. 'briggs', 'natural', 'uniform')",
    )
    advanced_image.add_argument(
        "--robust",
        type=float,
        default=0.0,
        help="Robust parameter for Briggs weighting (-2 to +2)",
    )
    advanced_image.add_argument(
        "--no_multiscale",
        action="store_false",
        dest="use_multiscale",
        help="Disable multiscale CLEAN for extended structures",
    )
    advanced_image.add_argument(
        "--clean_threshold",
        type=float,
        default=1.0,
        help="Clean threshold in sigma for final deconvolution (Note this is not auto-mask)",
    )
    advanced_image.add_argument(
        "--do_pbcor",
        action="store_true",
        help="Apply primary beam correction after imaging",
    )
    advanced_image.add_argument(
        "--no_apply_parang",
        action="store_false",
        dest="apply_parang",
        help="Disable parallactic angle rotation during imaging",
    )
    advanced_image.add_argument(
        "--cutout_rsun",
        type=float,
        default=2.5,
        help="Field of view cutout radius in solar radii",
    )
    advanced_image.add_argument(
        "--no_solar_mask",
        action="store_false",
        dest="use_solar_mask",
        help="Disable use solar disk mask during deconvolution",
    )
    advanced_image.add_argument(
        "--no_overlay",
        action="store_false",
        dest="make_overlay",
        help="Disable overlay plot on GOES SUVI after imaging",
    )

    # === Advanced options ===
    advanced = parser.add_argument_group(
        "###################\nAdvanced pipeline parameters\n###################"
    )
    advanced.add_argument(
        "--non_solar_data",
        action="store_false",
        dest="solar_data",
        help="Disable solar data mode",
    )
    advanced.add_argument(
        "--non_ds",
        action="store_false",
        dest="make_ds",
        help="Disable making solar dynamic spectra",
    )
    advanced.add_argument(
        "--do_forcereset_weightflag",
        action="store_true",
        help="Force reset of weights and flags (disabled by default)",
    )
    advanced.add_argument(
        "--no_noise_cal",
        action="store_false",
        dest="do_noise_cal",
        help="Disable noise calibration",
    )
    advanced.add_argument(
        "--no_cal_partition",
        action="store_false",
        dest="do_cal_partition",
        help="Disable calibrator MS partitioning",
    )
    advanced.add_argument(
        "--no_cal_flag",
        action="store_false",
        dest="do_cal_flag",
        help="Disable initial flagging of calibrators",
    )
    advanced.add_argument(
        "--no_import_model",
        action="store_false",
        dest="do_import_model",
        help="Disable model import",
    )
    advanced.add_argument(
        "--no_basic_cal",
        action="store_false",
        dest="do_basic_cal",
        help="Disable basic gain calibration",
    )
    advanced.add_argument(
        "--do_sidereal_cor",
        action="store_true",
        dest="do_sidereal_cor",
        help="Sidereal motion correction for Sun (disabled by default)",
    )
    advanced.add_argument(
        "--no_selfcal_split",
        action="store_false",
        dest="do_selfcal_split",
        help="Disable split for self-calibration",
    )
    advanced.add_argument(
        "--no_selfcal",
        action="store_false",
        dest="do_selfcal",
        help="Disable self-calibration",
    )
    advanced.add_argument(
        "--no_ap_selfcal",
        action="store_false",
        dest="do_ap_selfcal",
        help="Disable amplitude-phase self-calibration",
    )
    advanced.add_argument(
        "--no_solar_selfcal",
        action="store_false",
        dest="solar_selfcal",
        help="Disable solar-specific self-calibration parameters",
    )
    advanced.add_argument(
        "--no_target_split",
        action="store_false",
        dest="do_target_split",
        help="Disable target data split",
    )
    advanced.add_argument(
        "--no_applycal",
        action="store_false",
        dest="do_applycal",
        help="Disable application of basic calibration solutions",
    )
    advanced.add_argument(
        "--no_apply_selfcal",
        action="store_false",
        dest="do_apply_selfcal",
        help="Disable application of self-calibration solutions",
    )
    advanced.add_argument(
        "--no_imaging",
        action="store_false",
        dest="do_imaging",
        help="Disable final imaging",
    )

    # === Advanced local system/ per node hardware resource parameters ===
    advanced_resource = parser.add_argument_group(
        "###################\nAdvanced hardware resource parameters for local system or per node on HPC cluster\n###################"
    )
    advanced_resource.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="Fraction of CPU usuage per node",
    )
    advanced_resource.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Fraction of memory usuage per node",
    )
    advanced_resource.add_argument(
        "--keep_backup",
        action="store_true",
        help="Keep backup of intermediate steps",
    )
    advanced_resource.add_argument(
        "--no_remote_logger",
        action="store_false",
        dest="remote_logger",
        help="Disable remote logger",
    )
    
    
    # === Advanced local system/ per node hardware resource parameters ===
    advanced_hpc = parser.add_argument_group(
        "###################\nAdvanced HPC settings\n###################"
    )
    advanced_hpc.add_argument(
        "--n_nodes",
        type=int,
        default=0,
        help="Number of compute nodes to use (0 means local cluster)",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    result=prefect_server_status()
    if result is not True:
        print ("Prefect server is not running. First start it using init-meersolar-prefect start")
        sys.exit(1)
        
    try:
        print("\n########################################")
        print("Starting MeerSOLAR Pipeline....")
        print("#########################################\n")
        
        dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
            1,
            dask_dir=args.workdir,
            cpu_frac=args.cpu_frac,
            mem_frac=args.mem_frac,
        )
        npcu=int(psutil.cpu_count()*args.cpu_frac)
        dask_cluster.adapt(minimum=1, maximum=npcu) # 1 worker will be required to run the flow
        dask_addr=dask_client.scheduler.address
            
        jobid = get_jobid()    
        
        msg = master_control.with_options(flow_run_name=f"meersolar_{jobid}",task_runner=DaskTaskRunner(address=dask_addr))(
            msname=args.msname,
            workdir=args.workdir,
            outdir=args.outdir,
            solar_data=args.solar_data,
            # Pre-calibration
            do_forcereset_weightflag=args.do_forcereset_weightflag,
            do_cal_partition=args.do_cal_partition,
            do_cal_flag=args.do_cal_flag,
            do_import_model=args.do_import_model,
            # Basic calibration
            do_basic_cal=args.do_basic_cal,
            do_noise_cal=args.do_noise_cal,
            do_applycal=args.do_applycal,
            # Target data preparation
            do_target_split=args.do_target_split,
            target_scans=args.target_scans,
            freqrange=args.freqrange,
            timerange=args.timerange,
            uvrange=args.cal_uvrange,
            # Polarization calibration
            do_polcal=args.do_polcal,
            # Self-calibration
            do_selfcal=args.do_selfcal,
            do_selfcal_split=args.do_selfcal_split,
            do_apply_selfcal=args.do_apply_selfcal,
            do_ap_selfcal=args.do_ap_selfcal,
            solar_selfcal=args.solar_selfcal,
            solint=args.solint,
            # Sidereal correction
            do_sidereal_cor=args.do_sidereal_cor,
            # Dynamic spectra
            make_ds=args.make_ds,
            # Imaging
            do_imaging=args.do_imaging,
            do_pbcor=args.do_pbcor,
            weight=args.weight,
            robust=args.robust,
            minuv=args.minuv,
            image_freqres=args.image_freqres,
            image_timeres=args.image_timeres,
            pol=args.pol,
            apply_parang=args.apply_parang,
            clean_threshold=args.clean_threshold,
            use_multiscale=args.use_multiscale,
            use_solar_mask=args.use_solar_mask,
            cutout_rsun=args.cutout_rsun,
            make_overlay=args.make_overlay,
            # Resource settings
            cpu_frac=args.cpu_frac,
            mem_frac=args.mem_frac,
            n_nodes=args.n_nodes,
            keep_backup=args.keep_backup,
            # Remote logging
            remote_logger=args.remote_logger,
            dask_addr=dask_addr,
            jobid=jobid,
        )
    except Exception as e:
        traceback.print_exc()
    finally:
        drop_cache(args.msname)
        drop_cache(args.workdir)
        drop_cache(args.outdir)


if __name__ == "__main__":
    cli()



import logging
import dask
import numpy as np
import argparse
import traceback
import copy
import time
import sys
import os
from casatasks import casalog
from casatools import msmetadata
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def partion_ms(
    msname,
    outputms,
    workdir,
    fields="",
    scans="",
    width=1,
    timebin="",
    datacolumn="DATA",
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Perform mstransform of a single scan

    Parameters
    ----------
    msname : str
        Name of the measurement set
    outputms : str
        Output ms name
    workdir : str
        Work directory
    field : str, optional
        Fields to be splited
    scans : str, optional
        Scans to split
    width : int, optional
        Number of channels to average
    timebin : str, optional
        Time to average
    datacolumn : str, optional
        Data column to split
    ncpu : int, optional
        Number of CPU threads to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    str
        Output multi-measurement set name
    """
    print("##################")
    print("Paritioning measurement set: " + msname)
    print("##################\n")
    print("Determining valid scan list ....")
    start_time = time.time()
    valid_scans = get_valid_scans(msname, min_scan_time=1)
    msmd = msmetadata()
    msname = os.path.abspath(msname.rstrip("/"))
    msmd.open(msname)
    if scans != "":
        scan_list = scans.split(",")
    else:
        scan_list = msmd.scannumbers()
    scan_list = [int(i) for i in scan_list]
    if fields != "":  # Filtering scans only in the given fields
        scan_list_field = []
        field_list = []
        for i in fields.split(","):
            try:
                i = int(i)
            except BaseException:
                pass
            field_list.append(i)
        for field in field_list:
            a = msmd.scansforfield(field).tolist()
            scan_list_field = scan_list_field + a
        backup_scan_list = copy.deepcopy(scan_list)
        for s in scan_list:
            if s not in scan_list_field or s not in valid_scans:
                backup_scan_list.remove(s)
        scan_list = copy.deepcopy(backup_scan_list)
    else:
        backup_scan_list = copy.deepcopy(scan_list)
        for s in scan_list:
            if s not in valid_scans:
                backup_scan_list.remove(s)
        scan_list = copy.deepcopy(backup_scan_list)
    msmd.close()
    if len(scan_list) == 0:
        print("Please provide at-least one valid scan to split.")
        return

    field_list = []
    msmd = msmetadata()
    msmd.open(msname)
    field_names = msmd.fieldnames()
    for scan in scan_list:
        field = msmd.fieldsforscan(scan)[0]
        field_list.append(str(field_names[field]))
    msmd.close()
    msmd.done()
    field = ",".join(field_list)

    ###########################
    # Dask local cluster setup
    ###########################
    scan_sizes = []
    for scan in scan_list:
        scan_sizes.append(get_ms_scan_size(msname, int(scan)))
    total_required_size = round(2 * np.nansum(scan_sizes), 2)
    mem_limit = total_required_size / len(scan_list)
    if dask_addr is None:
        dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
            get_dask_client(
                len(scan_list),
                dask_dir=workdir,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                min_mem_per_job=mem_limit,
            )
        )
        dask_cluster.adapt(minimum=0, maximum=n_jobs)
    else:
        _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
            len(scan_list),
            dask_dir=workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            min_mem_per_job=mem_limit,
            only_cal=True,
        )
        os.system(f"rm -rf {dask_dir}")
        dask_client = Client(address=dask_addr)
    wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
    tasks = []
    for i in range(len(scan_list)):
        scan = scan_list[i]
        outputvis = f"{workdir}/scan_{scan}.ms"
        task = delayed(single_mstransform)(
            msname=msname,
            outputms=outputvis,
            scan=str(scan),
            field="",
            width=width,
            timebin=timebin,
            n_threads=n_threads,
            numsubms=1,
        )
        tasks.append(task)
    futures = dask_client.compute(tasks)
    splited_ms_list = list(dask_client.gather(futures))
    dask_client.close()
    if dask_addr is None:
        dask_cluster.close()
        os.system(f"rm -rf {dask_dir}")
    splited_ms_list_copy = copy.deepcopy(splited_ms_list)
    for ms in splited_ms_list:
        if ms is None:
            splited_ms_list_copy.remove(ms)
    splited_ms_list = copy.deepcopy(splited_ms_list_copy)
    outputms = outputms.rstrip("/")
    if os.path.exists(outputms):
        os.system("rm -rf " + outputms)
    if os.path.exists(outputms + ".flagversions"):
        os.system("rm -rf " + outputms + ".flagversions")
    if len(splited_ms_list) == 0:
        print("No splited ms to concat.")
    elif len(splited_ms_list) == 1:
        os.system(f"mv {splited_ms_list[0]} {outputms}")
    else:
        print("Making multi-MS ....")
        from casatasks import virtualconcat

        with suppress_casa_output():
            virtualconcat(vis=splited_ms_list, concatvis=outputms)
    print("##################")
    print("Total time taken : " + str(time.time() - start_time) + "s")
    print("##################\n")
    return outputms


def main(
    msname,
    outputms="multi.ms",
    workdir="",
    fields="",
    scans="",
    width=1,
    timebin="",
    datacolumn="data",
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid="0",
    start_remote_log=False,
    dask_addr=None,
):
    """
    Partition a measurement set using field, scan, channel, and time selection in subms.

    Parameters
    ----------
    msname : str
        Path to the input measurement set (MS).
    outputms : str, optional
        Name of the output measurement set. Default is "multi.ms".
    workdir : str, optional
        Directory for logs and intermediate files. If empty, defaults to `<msname>/workdir`.
    fields : str, optional
        Field selection string, in CASA syntax (e.g., "0~2"). Default is "" (all fields).
    scans : str, optional
        Scan selection string, comma-separated (e.g., "1,3,5"). Default is "" (all scans).
    width : int, optional
        Number of channels to average together spectrally. Default is 1 (no averaging).
    timebin : str, optional
        Time averaging interval (e.g., "10s", "1min"). Empty string disables time averaging. Default is "".
    datacolumn : str, optional
        Name of the data column to operate on (e.g., "data", "corrected"). Default is "data".
    cpu_frac : float, optional
        Fraction of available CPUs to use per task. Default is 0.8.
    mem_frac : float, optional
        Fraction of available memory to use per task. Default is 0.8.
    logfile : str or None, optional
        Path to log file. If None, logging to file is disabled. Default is None.
    jobid : str, optional
        Unique job identifier used for PID tracking. Default is "0".
    start_remote_log : bool, optional
        If True, enables remote logging using credentials in `workdir`. Default is False.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "partition_cal", logfile, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")
    ###########

    try:
        if os.path.exists(msname):
            outputms = partion_ms(
                msname,
                outputms,
                workdir,
                fields=fields,
                scans=scans,
                width=width,
                timebin=timebin,
                datacolumn=datacolumn,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
            if outputms is None or not os.path.exists(outputms):
                print("Error in partitioning measurement set.")
                msg = 1
            else:
                print("Partitioned multi-MS is created at:", outputms)
                msg = 0
        else:
            print("Please provide a valid measurement set.\n")
            msg = 1
    except Exception:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Partition measurement set in multi-MS format",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "msname",
        type=str,
        help="Name of input measurement set (required positional argument)",
    )
    basic_args.add_argument(
        "--outputms",
        type=str,
        default="multi.ms",
        help="Name of output multi-MS",
    )
    basic_args.add_argument("--workdir", type=str, required=True, help="Work directory")

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--fields",
        type=str,
        default="",
        help="Comma-separated list of field IDs to split",
    )
    adv_args.add_argument(
        "--scans",
        type=str,
        default="",
        help="Comma-separated list of scans to split",
    )
    adv_args.add_argument(
        "--width",
        type=int,
        default=1,
        help="Number of spectral channels to average",
    )
    adv_args.add_argument(
        "--timebin",
        type=str,
        default="",
        help="Time averaging bin (e.g., '10s', '1min')",
    )
    adv_args.add_argument(
        "--datacolumn",
        type=str,
        default="data",
        help="Datacolumn to split",
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="CPU fraction to use",
    )
    hard_args.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Memory fraction to use",
    )
    hard_args.add_argument("--logfile", type=str, default=None, help="Path to log file")
    hard_args.add_argument(
        "--jobid",
        type=str,
        default="0",
        help="Job ID for process tracking",
    )

    # Show help if nothing is passed
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        msname=args.msname,
        outputms=args.outputms,
        workdir=args.workdir,
        fields=args.fields,
        scans=args.scans,
        width=args.width,
        timebin=args.timebin,
        datacolumn=args.datacolumn,
        cpu_frac=float(args.cpu_frac),
        mem_frac=float(args.mem_frac),
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nMeasurement set partitioning is finished.\n###################\n"
    )
    sys.exit(result)



import resource
import logging
import psutil
import dask
import numpy as np
import argparse
import traceback
import copy
import time
import glob
import sys
import os
from casatasks import casalog
from casatools import msmetadata
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    logfile = casalog.logfile()
    os.system("rm -rf " + logfile)
except BaseException:
    pass

datadir = get_datadir()


def perform_imaging(
    msname="",
    workdir="",
    datacolumn="CORRECTED_DATA",
    freqrange="",
    timerange="",
    imagedir="",
    imsize=1024,
    cellsize=2,
    nchan=1,
    ntime=1,
    pol="I",
    weight="briggs",
    robust=0.0,
    minuv=0,
    threshold=1.0,
    use_multiscale=True,
    use_solar_mask=True,
    mask_radius=25,
    savemodel=True,
    saveres=True,
    ncpu=-1,
    mem=-1,
    band="",
    cutout_rsun=2.5,
    make_overlay=True,
    make_plots=True,
    logfile="imaging.log",
    dry_run=False,
):
    """
    Perform spectropolarimetric snapshot imaging of a ms

    Parameters
    ----------
    msname : str
        Name of the measurement set
    workdir : str
        Work directory name
    datacolumn : str, optional
        Data column
    freqrange : str, optional
        Frequency range to image
    imagedir : str, optional
        Image directory name (default: workdir). Images, models, residuals will be saved in directories named images. models, residuals inside imagedir
    imsize : int, optional
        Image size in pixels
    cellsize : float, optional
        Cell size in arcseconds
    nchan : int, optional
        Number of spectral channels
    ntime : int, optional
        Number of temporal slices
    pol : str, optional
        Stokes parameters to image
    weight : str, optional
        Image weighting scheme
    robust : float, optional
        Briggs weighting robustness parameter
    minuv : float, optional
        Minimum UV-lambda to be used in imaging
    threshold : float, optional
        CLEAN threshold
    use_multiscale : bool, optional
        Use multiscale or not
    use_solar_mask : bool, optional
        Use solar mask
    mask_radius : float, optional
        Mask radius in arcminute
    savemodel : bool, optional
        Save model images or not
    saveres : bool, optional
        Save residual images or not
    band : str, optional
        Band name
    cutout_rsun : float, optional
        Cutout image size in solar radii from center (default: 2.5 solar radii)
    make_overlay : bool, optional
        Make SUVI MeerKAT overlay
    make_plots : bool, optional
        Make radio map helioprojective plots
    logfile : str, optional
        Log file name
    ncpu : int, optional
        Number of CPU threads to use
    mem : float, optional
        Memory in GB to use

    Returns
    -------
    int
        Success message
    list
        List of images [[images],[models],[residuals]]
    """
    if os.path.exists(logfile):
        os.system(f"rm -rf {logfile}")
    if dry_run:
        process = psutil.Process(os.getpid())
        usemem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return usemem
    logger, logfile = create_logger(
        os.path.basename(logfile).split(".log")[0],
        logfile,
        verbose=False,
        get_print=True,
    )
    sub_observer = None
    if os.path.exists(f"{workdir}/jobname_password.npy") and logfile is not None:
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            sub_observer = init_logger(
                "remotelogger_imaging_{os.path.basename(msname).split('.ms')[0]}",
                logfile,
                jobname=jobname,
                password=password,
            )
    if os.path.exists(f"{workdir}/.attcal"):
        attcal = True
    elif os.path.exists(f"{workdir}/.noattcal"):
        attcal = False
    else:
        attcal = "NOINFO"
    try:
        msname = msname.rstrip("/")
        msname = os.path.abspath(msname)
        if band == "":
            band = get_band_name(msname)
        logger.info(f"{os.path.basename(msname)} --Perform imaging...\n")
        #########
        # Imaging
        #########
        msmd = msmetadata()
        msmd.open(msname)
        freq = msmd.meanfreq(0, unit="MHz")
        freqs = msmd.chanfreqs(0, unit="MHz")
        times = msmd.timesforspws(0)
        npol = msmd.ncorrforpol()[0]
        msmd.close()
        ###################################
        # Finding channel and time ranges
        ###################################
        if freqrange != "":
            start_chans = []
            end_chans = []
            freq_list = freqrange.split(",")
            for f in freq_list:
                start_freq = float(f.split("~")[0])
                end_freq = float(f.split("~")[-1])
                if start_freq >= np.nanmin(freqs) and end_freq <= np.nanmax(freqs):
                    start_chan = np.argmin(np.abs(start_freq - freqs))
                    end_chan = np.argmin(np.abs(end_freq - freqs))
                    start_chans.append(start_chan)
                    end_chans.append(end_chan)
        else:
            start_chans = [0]
            end_chans = [len(freqs)]
        if len(start_chans) == 0:
            print(f"Please provide valid channel range between 0 and {len(freqs)}")
            time.sleep(5)
            clean_shutdown(sub_observer)
            return 1, []
        if timerange != "":
            start_times = []
            end_times = []
            time_list = timerange.split(",")
            for timerange in time_list:
                start_time = timestamp_to_mjdsec(timerange.split("~")[0])
                end_time = timestamp_to_mjdsec(timerange.split("~")[-1])
                if start_time >= np.nanmin(times) and end_time <= np.nanmax(times):
                    start_times.append(np.argmin(np.abs(times - start_time)))
                    end_times.append(np.argmin(np.abs(times - end_time)))
        else:
            start_times = [0]
            end_times = [len(times)]
        if len(start_times) == 0:
            print(
                f"Please provide valid time range between {mjdsec_to_timestamp(times[0])} and {mjdsec_to_timestamp(times[-1])}"
            )
            time.sleep(5)
            clean_shutdown(sub_observer)
            return 1, []

        if npol < 4 and pol == "IQUV":
            pol = "I"
        if ncpu < 1:
            ncpu = psutil.cpu_count()
        if mem < 0:
            mem = psutil.virtual_memory().total / (1024**3)
        prefix = workdir + "/imaging_" + os.path.basename(msname).split(".ms")[0]
        if imagedir == "":
            imagedir = workdir
        os.makedirs(imagedir, exist_ok=True)
        if weight == "briggs":
            weight += " " + str(robust)
        if threshold <= 1:
            threshold = 1.1

        wsclean_args = [
            "-quiet",
            "-scale " + str(cellsize) + "asec",
            "-size " + str(imsize) + " " + str(imsize),
            "-no-dirty",
            "-gridder tuned-wgridder",
            "-weight " + weight,
            "-name " + prefix,
            "-pol " + str(pol),
            "-niter 10000",
            "-mgain 0.85",
            "-nmiter 5",
            "-gain 0.1",
            "-minuv-l " + str(minuv),
            "-j " + str(ncpu),
            "-abs-mem " + str(round(mem, 2)),
            "-auto-threshold 1 -auto-mask " + str(threshold),
            "-no-update-model-required",
        ]
        if datacolumn != "CORRECTED_DATA" and datacolumn != "corrected":
            wsclean_args.append("-data-column " + datacolumn)

        ngrid = int(ncpu / 2)
        if ngrid > 1:
            wsclean_args.append("-parallel-gridding " + str(ngrid))

        if pol == "I":
            wsclean_args.append("-no-negative")

        #####################################
        # Spectral imaging configuration
        #####################################
        if nchan > 1:
            wsclean_args.append(f"-channels-out {int(nchan)}")
            wsclean_args.append("-no-mf-weighting")
            wsclean_args.append("-join-channels")

        #####################################
        # Temporal imaging configuration
        #####################################
        if ntime > 1:
            wsclean_args.append(f"-intervals-out {int(ntime)}")

        ################################################
        # Creating and using a solar mask
        ################################################
        if use_solar_mask:
            fits_mask = prefix + "_solar-mask.fits"
            if os.path.exists(fits_mask) == False:
                logger.info(
                    f"{os.path.basename(msname)} -- Creating solar mask of size: {mask_radius} arcmin.\n",
                )
                fits_mask = create_circular_mask(
                    msname, cellsize, imsize, mask_radius=mask_radius
                )
            if fits_mask is not None and os.path.exists(fits_mask):
                wsclean_args.append("-fits-mask " + fits_mask)
        final_list_dic = {"image": [], "model": [], "residual": []}
        for i in range(len(start_chans)):
            for j in range(len(start_times)):
                temp_wsclean_args = copy.deepcopy(wsclean_args)
                temp_wsclean_args.append(
                    f"-channel-range {start_chans[i]} {end_chans[i]}"
                )
                temp_wsclean_args.append(f"-interval {start_times[j]} {end_times[j]}")

                ######################################
                # Multiscale configuration
                ######################################
                if use_multiscale:
                    num_pixel_in_psf = calc_npix_in_psf(weight, robust=robust)
                    chan_number = int((start_chans[i] + end_chans[i]) / 2)
                    freqMHz = freqs[chan_number]
                    sun_dia = calc_sun_dia(freqMHz)  # Sun diameter in arcmin
                    sun_rad = sun_dia / 2
                    multiscale_scales = calc_multiscale_scales(
                        msname,
                        num_pixel_in_psf,
                        chan_number=chan_number,
                        max_scale=sun_rad,
                    )
                    temp_wsclean_args.append("-multiscale")
                    temp_wsclean_args.append("-multiscale-gain 0.1")
                    temp_wsclean_args.append(
                        "-multiscale-scales "
                        + ",".join([str(s) for s in multiscale_scales])
                    )
                    mid_freq = np.nanmean(
                        freqs[int(start_chans[i]) : int(end_chans[i])]
                    )
                    scale_bias = get_multiscale_bias(mid_freq)
                    temp_wsclean_args.append(f"-multiscale-scale-bias {scale_bias}")
                    if imsize >= 2048 and 4 * max(multiscale_scales) < 1024:
                        temp_wsclean_args.append("-parallel-deconvolution 1024")
                elif imsize >= 2048:
                    temp_wsclean_args.append("-parallel-deconvolution 1024")

                ######################################
                # Running imaging
                ######################################
                wsclean_cmd = "wsclean " + " ".join(temp_wsclean_args) + " " + msname
                logger.info(
                    f"{os.path.basename(msname)} -- WSClean command: {wsclean_cmd}\n",
                )
                msg = run_wsclean(wsclean_cmd, "solarwsclean", verbose=False)
                if msg == 0:
                    os.system("rm -rf " + prefix + "*psf.fits")
                    ######################
                    # Making stokes cubes
                    ######################
                    pollist = [i.upper() for i in list(pol)]
                    if len(pollist) == 1:
                        imagelist = sorted(glob.glob(prefix + "*image.fits"))
                        if not savemodel:
                            os.system("rm -rf " + prefix + "*model.fits")
                        else:
                            modellist = sorted(glob.glob(prefix + "*model.fits"))
                        if not saveres:
                            os.system("rm -rf " + prefix + "*residual.fits")
                        else:
                            reslist = sorted(glob.glob(prefix + "*residual.fits"))
                    else:
                        imagelist = []
                        stokeslist = []
                        for p in pollist:
                            stokeslist.append(
                                sorted(glob.glob(prefix + "*" + p + "-image.fits"))
                            )
                        for i in range(len(stokeslist[0])):
                            wsclean_images = sorted(
                                [stokeslist[k][i] for k in range(len(pollist))]
                            )
                            image_prefix = os.path.basename(wsclean_images[0]).split(
                                "-image"
                            )[0]
                            image_cube = make_stokes_wsclean_imagecube(
                                wsclean_images,
                                image_prefix + f"_{pol}_image.fits",
                                keep_wsclean_images=False,
                            )
                            imagelist.append(image_cube)
                        del stokeslist
                        if not savemodel:
                            os.system("rm -rf " + prefix + "*model.fits")
                        else:
                            modellist = []
                            stokeslist = []
                            for p in pollist:
                                stokeslist.append(
                                    sorted(glob.glob(prefix + f"*{p}*model.fits"))
                                )
                            for i in range(len(stokeslist[0])):
                                wsclean_models = sorted(
                                    [stokeslist[k][i] for k in range(len(pollist))]
                                )
                                model_prefix = os.path.basename(
                                    wsclean_models[0]
                                ).split("-model")[0]
                                model_cube = make_stokes_wsclean_imagecube(
                                    wsclean_models,
                                    model_prefix + f"_{pol}_model.fits",
                                    keep_wsclean_images=False,
                                )
                                modellist.append(model_cube)
                            del stokeslist
                        if not saveres:
                            os.system("rm -rf " + prefix + "*residual.fits")
                        else:
                            reslist = []
                            stokeslist = []
                            for p in pollist:
                                stokeslist.append(
                                    sorted(glob.glob(prefix + f"*{p}*residual.fits"))
                                )
                            for i in range(len(stokeslist[0])):
                                wsclean_residuals = sorted(
                                    [stokeslist[k][i] for k in range(len(pollist))]
                                )
                                res_prefix = os.path.basename(
                                    wsclean_residuals[0]
                                ).split("-residual")[0]
                                residual_cube = make_stokes_wsclean_imagecube(
                                    wsclean_residuals,
                                    res_prefix + f"_{pol}_residual.fits",
                                    keep_wsclean_images=False,
                                )
                                reslist.append(residual_cube)
                            del stokeslist

                    ######################
                    # Renaming images
                    ######################
                    if len(imagelist) > 0:
                        logger.info(f"Total {len(imagelist)} images are made.")
                        logger.info("Renaming and making plots...")
                        os.makedirs(imagedir + "/images", exist_ok=True)
                        final_image_list = []
                        for imagename in imagelist:
                            renamed_image = rename_meersolar_image(
                                imagename,
                                imagedir=imagedir + "/images",
                                pol=pol,
                                band=band,
                                attcal=attcal,
                                cutout_rsun=cutout_rsun,
                                make_overlay=make_overlay,
                                make_plots=make_plots,
                            )
                            final_image_list.append(renamed_image)
                        final_list_dic["image"] = final_image_list
                        if savemodel and len(modellist) > 0:
                            final_model_list = []
                            os.makedirs(imagedir + "/models", exist_ok=True)
                            for modelname in modellist:
                                renamed_model = rename_meersolar_image(
                                    modelname,
                                    imagedir=imagedir + "/models",
                                    pol=pol,
                                    band=band,
                                    attcal=attcal,
                                    cutout_rsun=cutout_rsun,
                                    make_overlay=False,
                                    make_plots=False,
                                )
                                final_model_list.append(renamed_model)
                            final_list_dic["model"] = final_model_list
                        if saveres and len(reslist) > 0:
                            final_res_list = []
                            os.makedirs(imagedir + "/residuals", exist_ok=True)
                            for resname in reslist:
                                renamed_res = rename_meersolar_image(
                                    resname,
                                    imagedir=imagedir + "/residuals",
                                    pol=pol,
                                    band=band,
                                    attcal=attcal,
                                    cutout_rsun=cutout_rsun,
                                    make_overlay=False,
                                    make_plots=False,
                                )
                                final_res_list.append(renamed_res)
                            final_list_dic["residual"] = final_res_list
            if os.path.exists(f"{imagedir}/images/dask-scratch-space"):
                os.system(f"rm -rf {imagedir}/images/dask-scratch-space")
            if use_solar_mask and os.path.exists(fits_mask):
                os.system("rm -rf " + fits_mask)
            if len(final_list_dic["image"]) == 0:
                logger.info(
                    f"{os.path.basename(msname)} -- No image is made.\n",
                )
                time.sleep(5)
                clean_shutdown(sub_observer)
                return 1, final_list_dic
            else:
                logger.info(
                    f"{os.path.basename(msname)} -- Imaging is successfully done.\n",
                )
                time.sleep(5)
                clean_shutdown(sub_observer)
                return 0, final_list_dic
        else:
            if use_solar_mask and os.path.exists(fits_mask):
                os.system("rm -rf " + fits_mask)
            logger.info(
                f"{os.path.basename(msname)} -- No image is made.\n",
            )
            time.sleep(5)
            clean_shutdown(sub_observer)
            return 1, {}
    except Exception as e:
        traceback.print_exc()
        logger.info(
            f"{os.path.basename(msname)} -- Error in imaging.\n",
        )
        time.sleep(5)
        clean_shutdown(sub_observer)
        return 1, {}
    finally:
        time.sleep(5)
        drop_cache(msname)


def run_all_imaging(
    mslist=[],
    mainlogger=None,
    workdir="",
    outdir="",
    freqrange="",
    timerange="",
    datacolumn="CORRECTED_DATA",
    freqres=-1,
    timeres=-1,
    weight="briggs",
    robust=0.0,
    minuv=0,
    pol="I",
    threshold=1.0,
    use_multiscale=True,
    use_solar_mask=True,
    imaging_params={},  # TODO
    savemodel=False,
    saveres=False,
    band="",
    cutout_rsun=-1,
    make_overlay=True,
    make_plots=True,
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile="imaging.log",
    dask_addr=None,
):
    """
    Run spectropolarimetric snapshot imaging on a list of measurement sets

    Parameters
    ----------
    mslist : list
        Measurement set list
    mainlogger : str
        Python logger
    workdir : str
        Work directory
    outdir : str
        Output directory
    freqrange : str, optional
        Frequency range to image
    timerange : str, optional
        Time range
    datacolumn : str, optional
        Data column
    freqres : float, optional
        Frequency resolution of spectral chunk in MHz
    timeres : float, optional
        Time resolution of temporal chunk in seconds
    weight : str, optional
        Image weighting
    robust : float, optional
        Briggs weighting robust parameter
    minuv : float, optional
        Minimum UV-lambda to use in imaging
    pol : str, optional
        Stokes parameters to image
    threshold : float, optional
        CLEAN threshold
    use_multiscale : bool, optional
        Use multiscale or not
    use_solar_mask : bool, optional
        Use solar mask
    savemodel : bool, optional
        Save model images or not
    saveres : bool, optional
        Save residual images or not
    band : str, optional
        Band name
    cutout_rsun : float, optional
        Cutout image size (width ans height is : 2 times cutout_rsun)
        Default value: 2 solar radii
        Note: default FoV is 2 solar solar radii. If cutout_rsun is chosen larger than 2 solar radii, FoV will be increased accordingly.
    make_overlay : bool, optional
        Make SUVI MeerKAT overlay
    make_plots : bool, optional
        Make radio image helioprojective plots
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    start_time = time.time()
    mslist = sorted(mslist)
    observer = None
    if mainlogger is None:
        mainlog_file = workdir + "/logs/imaging_targets.log"
        mainlogger, mainlog_file = create_logger(
            os.path.basename(mainlog_file).split(".log")[0],
            mainlog_file,
            verbose=False,
            get_print=True,
        )
        observer = None
        if (
            os.path.exists(f"{workdir}/jobname_password.npy")
            and mainlog_file is not None
        ):
            time.sleep(5)
            jobname, password = np.load(
                f"{workdir}/jobname_password.npy", allow_pickle=True
            )
            if os.path.exists(mainlog_file):
                print(f"Starting remote logger. Remote logger password: {password}")
                observer = init_logger(
                    "all_imaging", mainlog_file, jobname=jobname, password=password
                )
    ###########################
    # WSClean container
    ###########################
    container_name = "solarwsclean"
    container_present = check_udocker_container(container_name)
    if not container_present:
        container_name = initialize_wsclean_container(name=container_name)
        if container_name is None:
            print(
                f"Container {container_name} is not initiated. First initiate container and then run."
            )
            return 1
    try:
        if len(mslist) == 0:
            mainlogger.error("Provide valid measurement set list.")
            time.sleep(5)
            clean_shutdown(observer)
            return 1
        if weight == "briggs":
            weight_str = f"{weight}_{robust}"
        else:
            weight_str = weight
        if freqres == -1 and timeres == -1:
            imagedir = outdir + f"/imagedir_f_all_t_all_w_{weight_str}"
        elif freqres != -1 and timeres == -1:
            imagedir = outdir + f"/imagedir_f_{freqres}_t_all_w_{weight_str}"
        elif freqres == -1 and timeres != -1:
            imagedir = outdir + f"/imagedir_f_all_t_{timeres}_w_{weight_str}"
        else:
            imagedir = outdir + f"/imagedir_f_{freqres}_t_{timeres}_w_{weight_str}"
        os.makedirs(imagedir, exist_ok=True)

        ####################################
        # Filtering any corrupted ms
        #####################################
        filtered_mslist = []  # Filtering in case any ms is corrupted
        for ms in mslist:
            checkcol = check_datacolumn_valid(ms)
            if checkcol:
                filtered_mslist.append(ms)
            else:
                mainlogger.warning(f"Issue in : {ms}")
                os.system(f"rm -rf {ms}")
        mslist = filtered_mslist

        #####################################
        # Determining spectro-temporal chunks
        #####################################
        if timeres < 0:
            ntime_list = [1] * len(mslist)
        else:
            ntime_list = []
            msmd = msmetadata()
            for ms in mslist:
                msmd.open(ms)
                times = msmd.timesforspws(0)
                msmd.close()
                tw = max(times) - min(times)
                ntime = max(1, int(tw / timeres))
                ntime_list.append(ntime)
        if freqres < 0:
            nchan_list = [1] * len(mslist)
        else:
            nchan_list = []
            msmd = msmetadata()
            for ms in mslist:
                msmd.open(ms)
                freqs = msmd.chanfreqs(0, unit="MHz")
                msmd.close()
                bw = max(freqs) - min(freqs)
                nchan = max(1, int(np.ceil(bw / freqres)))
                nchan_list.append(nchan)

        ######################################
        # Resetting maximum file limit
        ######################################
        soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
        new_soft_limit = max(soft_limit, int(0.8 * hard_limit))
        if soft_limit < new_soft_limit:
            resource.setrlimit(resource.RLIMIT_NOFILE, (new_soft_limit, hard_limit))
        total_fd = 0
        npol = len(pol)
        for i in range(len(mslist)):
            ms = mslist[i]
            nchan = nchan_list[i]
            ntime = ntime_list[i]
            per_job_fd = (
                npol * (nchan + 1) * ntime * 4 * 2
            )  # 4 types of images, 2 is fudge factor
            total_fd += per_job_fd
        if total_fd <= 0:
            total_fd = 1
        n_jobs = max(1, int(new_soft_limit / total_fd))
        n_jobs = min(len(mslist), n_jobs)

        #################################
        # Dask client setup
        #################################
        mem_limit = perform_imaging(dry_run=True)
        if dask_addr is None:
            dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    n_jobs,
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_cpu_per_job=3,
                    min_mem_per_job=mem_limit,
                )
            )
        else:
            _, _, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    n_jobs,
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_cpu_per_job=3,
                    min_mem_per_job=mem_limit,
                    only_cal=True,
                )
            )
            dask_client=Client(address=dask_addr)
            os.system(f"rm -rf {dask_dir}")
        wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
        tasks = []
        for i in range(len(mslist)):
            ms = mslist[i]
            nchan = nchan_list[i]
            ntime = ntime_list[i]
            num_pixel_in_psf = calc_npix_in_psf(weight, robust=robust)
            cellsize = calc_cellsize(ms, num_pixel_in_psf)
            instrument_fov = calc_field_of_view(ms, FWHM=False)
            msmd = msmetadata()
            msmd.open(ms)
            freqMHz = msmd.meanfreq(0, unit="MHz")
            msmd.close()
            sun_size = calc_sun_dia(freqMHz)
            fov = min(
                instrument_fov, 2 * sun_size * 60
            )  # 3 times sun size at that frequency
            if cutout_rsun == -1:
                cutout_rsun = 2 * round(
                    sun_size / 32, 2
                )  # Multiple of optical disk of the sun
            if fov < (2 * (cutout_rsun * 16) * 60):
                fov = 2 * (cutout_rsun * 16) * 60
            imsize = int(fov / cellsize)
            pow2 = np.ceil(np.log2(imsize)).astype("int")
            possible_sizes = []
            for p in range(pow2):
                for k in [3, 5]:
                    possible_sizes.append(k * 2**p)
            possible_sizes = np.sort(np.array(possible_sizes))
            possible_sizes = possible_sizes[possible_sizes >= imsize]
            imsize = max(1024, int(possible_sizes[0]))
            os.makedirs(workdir + "/logs", exist_ok=True)
            logfile = (
                workdir
                + "/logs/imaging_"
                + os.path.basename(ms).split(".ms")[0]
                + ".log"
            )
            mainlogger.info(
                f"Starting imaging for ms : {ms}, Log file : {logfile}\n",
            )
            tasks.append(
                delayed(perform_imaging)(
                    msname=ms,
                    workdir=workdir,
                    freqrange=freqrange,
                    timerange=timerange,
                    datacolumn=datacolumn,
                    imagedir=imagedir,
                    imsize=imsize,
                    cellsize=cellsize,
                    nchan=nchan,
                    ntime=ntime,
                    pol=pol,
                    weight=weight,
                    robust=robust,
                    minuv=minuv,
                    threshold=threshold,
                    use_multiscale=use_multiscale,
                    use_solar_mask=use_solar_mask,
                    savemodel=savemodel,
                    saveres=saveres,
                    band=band,
                    cutout_rsun=cutout_rsun,
                    make_overlay=make_overlay,
                    make_plots=make_plots,
                    ncpu=n_threads,
                    mem=mem_limit,
                    logfile=logfile,
                )
            )
        futures = dask_client.compute(tasks)
        results = list(dask_client.gather(futures))
        dask_client.close()
        if dask_addr is None:
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
        all_image_list = []
        all_imaged_ms_list = []
        for i in range(len(results)):
            r = results[i][1]
            if len(r) == 0:
                mainlogger.info(
                    f"Imaging failed for ms : {mslist[i]}",
                )
            else:
                image_list = r["image"]
                if len(image_list) == 0:
                    mainlogger.info(
                        f"No image is made for ms : {mslist[i]}",
                    )
                else:
                    all_imaged_ms_list.append(mslist[i])
                    for image in image_list:
                        all_image_list.append(image)
        mainlogger.info(
            f"Numbers of input measurement sets : {len(mslist)}.",
        )
        mainlogger.info(
            f"Imaging successfully done for: {len(all_imaged_ms_list)} measurement sets.",
        )
        mainlogger.info(f"Total images made: {len(all_image_list)}.")
        mainlogger.info(
            f"Total time taken: {round(time.time()-start_time,2)}s",
        )
        time.sleep(5)
        clean_shutdown(observer)
        return 0
    except Exception as e:
        traceback.print_exc()
        mainlogger.info(
            f"Total time taken: {round(time.time()-start_time,2)}s",
        )
        time.sleep(5)
        clean_shutdown(observer)
        return 1


def main(
    mslist,
    workdir,
    outdir,
    freqrange="",
    timerange="",
    datacolumn="CORRECTED_DATA",
    pol="I",
    freqres=-1,
    timeres=-1,
    weight="briggs",
    robust=0.0,
    minuv=0.0,
    threshold=1.0,
    band="",
    cutout_rsun=-1,
    use_multiscale=True,
    use_solar_mask=True,
    savemodel=True,
    saveres=True,
    make_overlay=True,
    make_plots=True,
    start_remote_log=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    jobid=0,
    dask_addr=None,
):
    """
    Perform distributed spectropolarimetric snapshot imaging on multiple measurement sets.

    Parameters
    ----------
    mslist : str
        Comma-separated list of measurement set paths to be imaged.
    workdir : str
        Directory for intermediate files, logs, and temporary outputs.
    outdir : str
        Directory where final images, models, and plots will be saved.
    freqrange : str, optional
        Frequency range to image (e.g., "500~1000MHz"). Default is "" (all frequencies).
    timerange : str, optional
        Time range to image (e.g., "09:30:00~09:40:00"). Default is "" (all times).
    datacolumn : str, optional
        Data column to image (e.g., "DATA", "CORRECTED_DATA"). Default is "CORRECTED_DATA".
    pol : str, optional
        Polarization product to image (e.g., "I", "XX", "RR", "QUV"). Default is "I".
    freqres : float, optional
        Frequency resolution in MHz for slicing the MS. Use -1 to disable. Default is -1.
    timeres : float, optional
        Time resolution in seconds for snapshot imaging. Use -1 to disable. Default is -1.
    weight : str, optional
        Weighting scheme for imaging ("natural", "uniform", "briggs"). Default is "briggs".
    robust : float, optional
        Robustness parameter for Briggs weighting. Default is 0.0.
    minuv : float, optional
        Minimum uv-distance (in wavelengths) to include in imaging. Default is 0.0.
    threshold : float, optional
        Cleaning threshold in Jy. Default is 1.0.
    band : str, optional
        Optional label for the frequency band used in filenames and plots. Default is "".
    cutout_rsun : float, optional
        Radius in solar radii to cut out around solar center. Set to -1 to disable. Default is -1.
    use_multiscale : bool, optional
        If True, enables multiscale CLEAN deconvolution. Default is True.
    use_solar_mask : bool, optional
        If True, applies a solar disk mask during CLEAN to reduce sidelobe artifacts. Default is True.
    savemodel : bool, optional
        If True, saves the CLEAN model images. Default is True.
    saveres : bool, optional
        If True, saves the residual images. Default is True.
    make_overlay : bool, optional
        If True, generates image overlays on solar maps. Default is True.
    make_plots : bool, optional
        If True, generates diagnostic plots for each image. Default is True.
    start_remote_log : bool, optional
        Whether to enable remote logging using credentials in the workdir. Default is False.
    cpu_frac : float, optional
        Fraction of total CPUs to use per task. Default is 0.8.
    mem_frac : float, optional
        Fraction of total system memory to use per task. Default is 0.8.
    jobid : int, optional
        Unique job identifier for logging and PID tracking. Default is 0.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")
    mslist = mslist.split(",")

    if workdir == "":
        first_ms = mslist[0]
        workdir = os.path.dirname(os.path.abspath(first_ms)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    if outdir == "" or not os.path.exists(outdir):
        outdir = workdir
    os.makedirs(outdir, exist_ok=True)

    os.makedirs(workdir + "/logs/", exist_ok=True)

    mainlog_file = workdir + "/logs/imaging_targets.log"
    mainlogger, mainlog_file = create_logger(
        os.path.basename(mainlog_file).split(".log")[0],
        mainlog_file,
        get_print=True,
        verbose=False,
    )

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and mainlog_file is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(mainlog_file):
            observer = init_logger(
                "all_imaging", mainlog_file, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")
    ##########

    try:
        if len(mslist) == 0:
            mainlogger.info("Please provide a valid measurement set list.")
            msg = 1
        else:
            msg = run_all_imaging(
                mslist=mslist,
                mainlogger=mainlogger,
                workdir=workdir,
                outdir=outdir,
                freqrange=freqrange,
                timerange=timerange,
                datacolumn=datacolumn,
                freqres=freqres,
                timeres=timeres,
                weight=weight,
                robust=robust,
                minuv=minuv,
                threshold=threshold,
                use_multiscale=use_multiscale,
                use_solar_mask=use_solar_mask,
                pol=pol,
                band=band,
                make_plots=make_plots,
                cutout_rsun=cutout_rsun,
                make_overlay=make_overlay,
                savemodel=savemodel,
                saveres=saveres,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
    except Exception:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        for ms in mslist:
            drop_cache(ms)
        drop_cache(workdir)
        drop_cache(outdir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Perform spectropolarimetric snapshot imaging",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "mslist",
        type=str,
        help="Comma-separated list of measurement sets (required)",
    )
    basic_args.add_argument(
        "--workdir",
        type=str,
        required=True,
        default="",
        help="Work directory for imaging",
    )
    basic_args.add_argument(
        "--outdir",
        type=str,
        required=True,
        default="",
        help="Output directory for imaging products",
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced imaging parameters\n###################"
    )
    adv_args.add_argument(
        "--freqrange",
        type=str,
        default="",
        help="Frequency range to image",
    )
    adv_args.add_argument(
        "--timerange",
        type=str,
        default="",
        help="Time range to image",
    )
    adv_args.add_argument(
        "--datacolumn",
        type=str,
        default="CORRECTED_DATA",
        help="Data column to use for imaging",
    )
    adv_args.add_argument(
        "--pol",
        type=str,
        default="I",
        help="Stokes parameters to image",
    )
    adv_args.add_argument(
        "--freqres",
        type=float,
        default=-1,
        help="Frequency resolution per chunk in MHz (-1 for full)",
    )
    adv_args.add_argument(
        "--timeres",
        type=float,
        default=-1,
        help="Time resolution per chunk in seconds (-1 for full)",
    )
    adv_args.add_argument(
        "--weight",
        type=str,
        default="briggs",
        help="Imaging weighting scheme",
    )
    adv_args.add_argument(
        "--robust",
        type=float,
        default=0.0,
        help="Briggs robust parameter",
    )
    adv_args.add_argument(
        "--minuv_l",
        dest="minuv",
        type=float,
        default=0.0,
        help="Minimum UV distance in wavelengths",
    )
    adv_args.add_argument(
        "--threshold",
        type=float,
        default=1.0,
        help="CLEAN threshold in Jy",
    )
    adv_args.add_argument(
        "--band",
        type=str,
        default="",
        help="Band name label for output",
    )
    adv_args.add_argument(
        "--cutout_rsun",
        type=float,
        default=-1,
        help="Cutout radius for images (solar radii)",
    )
    adv_args.add_argument(
        "--no_multiscale",
        action="store_false",
        dest="use_multiscale",
        help="Do not use multiscale CLEAN",
    )
    adv_args.add_argument(
        "--no_solar_mask",
        action="store_false",
        dest="use_solar_mask",
        help="Do not use solar disk mask for CLEANing",
    )
    adv_args.add_argument(
        "--no_savemodel",
        action="store_false",
        dest="savemodel",
        help="Do no save model images",
    )
    adv_args.add_argument(
        "--no_saveres",
        action="store_false",
        dest="saveres",
        help="Do not save residual images",
    )
    adv_args.add_argument(
        "--no_make_overlay",
        action="store_false",
        dest="make_overlay",
        help="Do not generate overlay with SUVI images",
    )
    adv_args.add_argument(
        "--no_make_plots",
        action="store_false",
        dest="make_plots",
        help="Do not make generate helioprojective plots",
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="Fraction of available CPU to use",
    )
    hard_args.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Fraction of available memory to use",
    )
    hard_args.add_argument(
        "--jobid",
        type=str,
        default="0",
        help="Job ID for process tracking and logging",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        mslist=args.mslist,
        workdir=args.workdir,
        outdir=args.outdir,
        freqrange=args.freqrange,
        timerange=args.timerange,
        datacolumn=args.datacolumn,
        pol=args.pol,
        freqres=args.freqres,
        timeres=args.timeres,
        weight=args.weight,
        robust=args.robust,
        minuv=args.minuv,
        threshold=args.threshold,
        band=args.band,
        cutout_rsun=args.cutout_rsun,
        use_multiscale=args.use_multiscale,
        use_solar_mask=args.use_solar_mask,
        savemodel=args.savemodel,
        saveres=args.saveres,
        make_overlay=args.make_overlay,
        make_plots=args.make_plots,
        start_remote_log=args.start_remote_log,
        cpu_frac=float(args.cpu_frac),
        mem_frac=float(args.mem_frac),
        jobid=args.jobid,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    if result > 0:
        result = 1
    print("\n###################\nImaging is done.\n###################\n")
    os._exit(result)



import types
import psutil
import numpy as np
import traceback
import glob
import os
from datetime import datetime as dt, timezone
from .basic_utils import *
from .resource_utils import *

###############################
# Flagging related functions
################################


def do_flag_backup(msname, flagtype="flagdata"):
    """
    Take a flag backup

    Parameters
    ----------
    msname : str
        Measurement set name
    flagtype : str, optional
        Flag type
    """
    from casatools import agentflagger

    af = agentflagger()
    af.open(msname)
    versionlist = af.getflagversionlist()
    if len(versionlist) != 0:
        for version_name in versionlist:
            if flagtype in version_name:
                try:
                    version_num = (
                        int(version_name.split(":")[0].split(" ")[0].split("_")[-1]) + 1
                    )
                except BaseException:
                    version_num = 1
            else:
                version_num = 1
    else:
        version_num = 1
    dt_string = dt.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    af.saveflagversion(
        flagtype + "_" + str(version_num), "Flags autosave on " + dt_string
    )
    af.done()


def get_unflagged_antennas(msname="", scan="", n_threads=-1, dry_run=False):
    """
    Get unflagged antennas of a scan

    Parameters
    ----------
    msname : str
        Name of the measurement set
    scan : str
        Scans

    Returns
    -------
    numpy.array
        Unflagged antenna names
    numpy.array
        Flag fraction list
    """
    limit_threads(n_threads=n_threads)
    from casatasks import flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    mspath = os.path.dirname(os.path.abspath(msname))
    os.chdir(mspath)
    with suppress_casa_output():
        flag_summary = flagdata(vis=msname, scan=str(scan), mode="summary")
    antenna_flags = flag_summary["antenna"]
    unflagged_antenna_names = []
    flag_frac_list = []
    for ant in antenna_flags.keys():
        flag_frac = antenna_flags[ant]["flagged"] / antenna_flags[ant]["total"]
        if flag_frac < 1.0:
            unflagged_antenna_names.append(ant)
            flag_frac_list.append(flag_frac)
    unflagged_antenna_names = np.array(unflagged_antenna_names)
    flag_frac_list = np.array(flag_frac_list)
    return unflagged_antenna_names, flag_frac_list


def get_chans_flag(msname="", field="", n_threads=-1, dry_run=False):
    """
    Get flag/unflag channel list

    Parameters
    ----------
    msname : str
        Measurement set name
    field : str, optional
        Field name or ID

    Returns
    -------
    list
        Unflag channel list
    list
        Flag channel list
    """
    limit_threads(n_threads=n_threads)
    from casatasks import flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    mspath = os.path.dirname(os.path.abspath(msname))
    os.chdir(mspath)
    with suppress_casa_output():
        summary = flagdata(vis=msname, field=field, mode="summary", spwchan=True)
    unflag_chans = []
    flag_chans = []
    for chan in summary["spw:channel"]:
        r = summary["spw:channel"][chan]
        chan_number = int(chan.split("0:")[-1])
        flag_frac = r["flagged"] / r["total"]
        if flag_frac == 1:
            flag_chans.append(chan_number)
        else:
            unflag_chans.append(chan_number)
    return unflag_chans, flag_chans


def calc_flag_fraction(msname="", field="", scan="", n_threads=-1, dry_run=False):
    """
    Function to calculate the fraction of total data flagged.


    Parameters
    ----------
    msname : str
        Name of the measurement set
    field : str, optional
        Field names
    scan : str, optional
        Scan names

    Returns
    -------
    float
        Fraction of the total data flagged
    """
    limit_threads(n_threads=n_threads)
    from casatasks import flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    mspath = os.path.dirname(os.path.abspath(msname))
    os.chdir(mspath)
    with suppress_casa_output():
        summary = flagdata(vis=msname, field=field, scan=scan, mode="summary")
    flagged_fraction = summary["flagged"] / summary["total"]
    return flagged_fraction


def flag_outside_uvrange(vis, uvrange, n_threads=-1, flagbackup=True):
    """
    Flag outside the given uv range

    Parameters
    ----------
    vis : str
        Measurement set name
    uvrange : str
        UV-range
    flagbackup : bool, optional
        Flag backup
    """
    limit_threads(n_threads=n_threads)
    from casatasks import flagdata

    try:
        if "lambda" in uvrange:
            islambda = True
            uvrange = uvrange.replace("lambda", "")
        else:
            islambda = False
        if "~" in uvrange:
            low, high = uvrange.split("~")
            if islambda:
                low = f"{low}lambda"
                high = f"{high}lambda"
            cmds = [
                {"mode": "manual", "uvrange": f"<{low}", "flagbackup": flagbackup},
                {"mode": "manual", "uvrange": f">{high}", "flagbackup": flagbackup},
            ]
        elif ">" in uvrange:
            low = uvrange.split(">")[-1]
            if islambda:
                low = f"{low}lambda"
            cmds = [
                {"mode": "manual", "uvrange": f"<{low}", "flagbackup": flagbackup},
            ]
        elif "<" in uvrange:
            high = uvrange.split("<")[-1]
            if islambda:
                high = f"{high}lambda"
            cmds = [
                {"mode": "manual", "uvrange": f">{high}", "flagbackup": flagbackup},
            ]
        else:
            cmds = []
        if len(cmds) > 0:
            for cmd in cmds:
                print(f"Flagging command: {cmd}")
                flagdata(vis=vis, **cmd)
        return 0
    except Exception as e:
        traceback.print_exc()
        return 1


# Expose functions and classes
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]
